# -*- coding: utf-8 -*-

from DisplayCAL.dev.mocks import check_call, check_call_str
from tests.data.display_data import DisplayData


def test_update_estimated_measurement_time_1(argyll):
    """Testing for issue #37

    wxReportFrame.ReportFrame.update_estimated_measurement_time() method raising
    TypeError.
    """
    from DisplayCAL.config import initcfg
    from DisplayCAL.wxReportFrame import ReportFrame
    import wx

    initcfg()
    app = wx.GetApp() or wx.App()

    with check_call_str(
        "DisplayCAL.worker.Worker.get_instrument_name",
        "i1 DisplayPro, ColorMunki Display",
        call_count=2,
    ):
        report_frame = ReportFrame()
        # this shouldn't raise any TypeErrors as reported in #37
        report_frame.update_estimated_measurement_time("chart")



#!/usr/bin/env python3
#
# Copyright (C) 2009, Florian Hoech
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, 5th Floor, Boston, MA 02110-1301, USA

# DEV NOTES
#
# Currently not implemented in the Manifest class:
# * Validation (only very basic sanity checks are currently in place)
# * comClass, typelib, comInterfaceProxyStub and windowClass child elements of
#   the file element
# * comInterfaceExternalProxyStub and windowClass child elements of the
#   assembly element
# * Application Configuration File and Multilanguage User Interface (MUI)
#   support when searching for assembly files
#
# Isolated Applications and Side-by-side Assemblies:
# http://msdn.microsoft.com/en-us/library/dd408052%28VS.85%29.aspx
#
# Changelog:
# 2009-12-17  fix: small glitch in toxml / toprettyxml methods (xml declaration
#                  wasn't replaced when a different encodig than UTF-8 was used)
#             chg: catch xml.parsers.expat.ExpatError and re-raise as
#                  ManifestXMLParseError
#             chg: support initialize option in parse method also
#
# 2009-12-13  fix: fixed os import
#             fix: skip invalid / empty dependent assemblies
#
# 2009-08-21  fix: Corrected assembly searching sequence for localized
#                  assemblies
#             fix: Allow assemblies with no dependent files
#
# 2009-07-31  chg: Find private assemblies even if unversioned
#             add: Manifest.same_id method to check if two manifests have the
#                  same assemblyIdentity
#
# 2009-07-30  fix: Potential failure in File.calc_hash method if hash
#                  algorythm not supported
#             add: Publisher configuration (policy) support when searching for
#                  assembly files
#             fix: Private assemblies are now actually found if present (and no
#                  shared assembly exists)
#             add: Python 2.3 compatibility (oldest version supported by
#                  pyinstaller)
#
# 2009-07-28  chg: Code cleanup, removed a bit of redundancy
#             add: silent mode (set silent attribute on module)
#             chg: Do not print messages in silent mode
#
# 2009-06-18  chg: Use glob instead of regular expression in Manifest.find_files
#
# 2009-05-04  fix: Don't fail if manifest has empty description
#             fix: Manifests created by the toxml, toprettyxml, writexml or
#                  writeprettyxml methods are now correctly recognized by
#                  Windows, which expects the XML declaration to be ordered
#                  version-encoding-standalone (standalone being optional)
#             add: 'encoding' keyword argument in toxml, toprettyxml, writexml
#                  and writeprettyxml methods
#             chg: UpdateManifestResourcesFromXML and
#                  UpdateManifestResourcesFromXMLFile: set resource name
#                  depending on file type ie. exe or dll
#             fix: typo in __main__: UpdateManifestResourcesFromDataFile
#                  should have been UpdateManifestResourcesFromXMLFile
#
# 2009-03-21  First version

"""
winmanifest.py

Create, parse and write MS Windows Manifest files.
Find files which are part of an assembly, by searching shared and
private assemblies.
Update or add manifest resources in Win32 PE files.

Commandline usage:
winmanifest.py <dstpath> <xmlpath>
Updates or adds manifest <xmlpath> as resource in Win32 PE file <dstpath>.

"""

try:
    import hashlib
except ImportError:
    hashlib = None
    import md5
    import sha
import os
from glob import glob
import sys
import xml
from xml.dom import Node, minidom
from xml.dom.minidom import Document, Element

try:
    import winresource
except ImportError as detail:
    winresource = None
    print("W:", detail)
    print("W: Cannot check for assembly dependencies - resource access ")
    print("W: unavailable. To enable resource access, please install ")
    print("W: http://sourceforge.net/projects/pywin32/")

silent = False  # True suppresses all messages

LANGUAGE_NEUTRAL_NT5 = "x-ww"
LANGUAGE_NEUTRAL_NT6 = "none"
RT_MANIFEST = 24

Document.aChild = Document.appendChild
Document.cE = Document.createElement
Document.cT = Document.createTextNode
Document.getEByTN = Document.getElementsByTagName
Element.aChild = Element.appendChild
Element.getA = Element.getAttribute
Element.getEByTN = Element.getElementsByTagName
Element.remA = Element.removeAttribute
Element.setA = Element.setAttribute


def getChildElementsByTagName(self, tagName):
    """Return child elements of type tagName if found, else []"""
    result = []
    for child in self.childNodes:
        if isinstance(child, Element):
            if child.tagName == tagName:
                result.append(child)
    return result


def getFirstChildElementByTagName(self, tagName):
    """Return the first element of type tagName if found, else None"""
    for child in self.childNodes:
        if isinstance(child, Element):
            if child.tagName == tagName:
                return child
    return None


Document.getCEByTN = getChildElementsByTagName
Document.getFCEByTN = getFirstChildElementByTagName
Element.getCEByTN = getChildElementsByTagName
Element.getFCEByTN = getFirstChildElementByTagName


class _Hash(object):
    def __init__(self):
        self.md5 = md5.new
        self.sha = sha.new


if hashlib is None:
    hashlib = _Hash()


class _Dummy:
    pass


if winresource:
    _File = winresource.File
else:
    _File = _Dummy


class File(_File):

    """A file referenced by an assembly inside a manifest."""

    def __init__(
        self,
        filename="",
        hashalg=None,
        hash=None,
        comClasses=None,
        typelibs=None,
        comInterfaceProxyStubs=None,
        windowClasses=None,
    ):
        if winresource:
            winresource.File.__init__(self, filename)
        else:
            self.filename = filename
        self.name = os.path.basename(filename)
        if hashalg:
            self.hashalg = hashalg.upper()
        else:
            self.hashalg = None
        if (
            os.path.isfile(filename)
            and hashalg
            and hashlib
            and hasattr(hashlib, hashalg.lower())
        ):
            self.calc_hash()
        else:
            self.hash = hash
        self.comClasses = comClasses or []  # TO-DO: implement
        self.typelibs = typelibs or []  # TO-DO: implement
        self.comInterfaceProxyStubs = comInterfaceProxyStubs or []  # TO-DO: implement
        self.windowClasses = windowClasses or []  # TO-DO: implement

    def calc_hash(self, hashalg=None):
        """Calculate the hash of the file.

        Will be called automatically from the constructor if the file exists
        and hashalg is given (and supported), but may also be called manually
        e.g. to update the hash if the file has changed.

        """
        fd = open(self.filename, "rb")
        buf = fd.read()
        fd.close()
        if hashalg:
            self.hashalg = hashalg.upper()
        self.hash = getattr(hashlib, self.hashalg.lower())(buf).hexdigest()

    def find(self, searchpath):
        if not silent:
            print("I: Searching for file", self.name)
        fn = os.path.join(searchpath, self.name)
        if os.path.isfile(fn):
            if not silent:
                print("I: Found file", fn)
            return fn
        else:
            if not silent:
                print("W: No such file", fn)
            return None


class InvalidManifestError(Exception):
    pass


class ManifestXMLParseError(InvalidManifestError):
    pass


class Manifest(object):

    # Manifests:
    # http://msdn.microsoft.com/en-us/library/aa375365%28VS.85%29.aspx

    """
    Manifest constructor.

    To build a basic manifest for your application:
      mf = Manifest(type='win32', name='YourAppName', language='*',
                    processorArchitecture='x86', version=[1, 0, 0, 0])

    To write the XML to a manifest file:
      mf.writexml("YourAppName.exe.manifest")
    or
      mf.writeprettyxml("YourAppName.exe.manifest")

    """

    def __init__(
        self,
        manifestVersion=None,
        noInheritable=False,
        noInherit=False,
        type_=None,
        name=None,
        language=None,
        processorArchitecture=None,
        version=None,
        publicKeyToken=None,
        description=None,
        requestedExecutionLevel=None,
        uiAccess=None,
        dependentAssemblies=None,
        files=None,
        comInterfaceExternalProxyStubs=None,
    ):
        self.filename = None
        self.optional = None
        self.manifestType = "assembly"
        self.manifestVersion = manifestVersion or [1, 0]
        self.noInheritable = noInheritable
        self.noInherit = noInherit
        self.type = type_
        self.name = name
        self.language = language
        self.processorArchitecture = processorArchitecture
        self.version = version
        self.publicKeyToken = publicKeyToken
        # publicKeyToken:
        # A 16-character hexadecimal string that represents the last 8 bytes
        # of the SHA-1 hash of the public key under which the assembly is
        # signed. The public key used to sign the catalog must be 2048 bits
        # or greater. Required for all shared side-by-side assemblies.
        # http://msdn.microsoft.com/en-us/library/aa375692(VS.85).aspx
        self.applyPublisherPolicy = None
        self.description = None
        self.requestedExecutionLevel = requestedExecutionLevel
        self.uiAccess = uiAccess
        self.dependentAssemblies = dependentAssemblies or []
        self.bindingRedirects = []
        self.files = files or []
        self.comInterfaceExternalProxyStubs = (
            comInterfaceExternalProxyStubs or []
        )  # TO-DO: implement

    def add_dependent_assembly(
        self,
        manifestVersion=None,
        noInheritable=False,
        noInherit=False,
        type_=None,
        name=None,
        language=None,
        processorArchitecture=None,
        version=None,
        publicKeyToken=None,
        description=None,
        requestedExecutionLevel=None,
        uiAccess=None,
        dependentAssemblies=None,
        files=None,
        comInterfaceExternalProxyStubs=None,
    ):
        """Shortcut for self.dependentAssemblies.append(Manifest(*args, **kwargs))"""
        self.dependentAssemblies.append(
            Manifest(
                manifestVersion,
                noInheritable,
                noInherit,
                type_,
                name,
                language,
                processorArchitecture,
                version,
                publicKeyToken,
                description,
                requestedExecutionLevel,
                uiAccess,
                dependentAssemblies,
                files,
                comInterfaceExternalProxyStubs,
            )
        )
        if self.filename:
            # Enable search for private assembly by assigning bogus filename
            # (only the directory has to be correct)
            self.dependentAssemblies[-1].filename = ":".join((self.filename, name))

    def add_file(
        self,
        name="",
        hashalg="",
        hash="",
        comClasses=None,
        typelibs=None,
        comInterfaceProxyStubs=None,
        windowClasses=None,
    ):
        """Shortcut for manifest.files.append"""
        self.files.append(
            File(
                name,
                hashalg,
                hash,
                comClasses,
                typelibs,
                comInterfaceProxyStubs,
                windowClasses,
            )
        )

    def find_files(self, ignore_policies=True):
        """Search shared and private assemblies and return a list of files.

        If any files are not found, return an empty list.

        IMPORTANT NOTE: For the purpose of getting the dependent assembly
        files of an executable, the publisher configuration (aka policy)
        should be ignored (which is the default). Setting ignore_policies=False
        is only useful to find out which files are actually loaded at
        runtime.

        """

        # Shared Assemblies:
        # http://msdn.microsoft.com/en-us/library/aa375996%28VS.85%29.aspx
        #
        # Private Assemblies:
        # http://msdn.microsoft.com/en-us/library/aa375674%28VS.85%29.aspx
        #
        # Assembly Searching Sequence:
        # http://msdn.microsoft.com/en-us/library/aa374224%28VS.85%29.aspx
        #
        # NOTE:
        # Multilanguage User Interface (MUI) support not yet implemented

        files = []

        languages = []
        if self.language not in (None, "", "*", "neutral"):
            languages.append(self.getlanguage())
            if "-" in self.language:
                # language-culture syntax, e.g. en-us
                # Add only the language part
                languages.append(self.language.split("-")[0])
            if self.language not in ("en-us", "en"):
                languages.append("en-us")
            if self.language != "en":
                languages.append("en")
        languages.append(self.getlanguage("*"))

        winsxs = os.path.join(os.getenv("SystemRoot"), "WinSxS")
        if not os.path.isdir(winsxs) and not silent:
            print("W: No such dir", winsxs)
        manifests = os.path.join(winsxs, "Manifests")
        if not os.path.isdir(manifests) and not silent:
            print("W: No such dir", manifests)
        if not ignore_policies and self.version:
            if sys.getwindowsversion() < (6,):
                # Windows XP
                pcfiles = os.path.join(winsxs, "Policies")
                if not os.path.isdir(pcfiles) and not silent:
                    print("W: No such dir", pcfiles)
            else:
                # Vista or later
                pcfiles = manifests

        for language in languages:
            version = self.version

            # Search for publisher configuration
            if not ignore_policies and version:
                # Publisher Configuration (aka policy)
                # A publisher configuration file globally redirects
                # applications and assemblies having a dependence on one
                # version of a side-by-side assembly to use another version of
                # the same assembly. This enables applications and assemblies
                # to use the updated assembly without having to rebuild all of
                # the affected applications.
                # http://msdn.microsoft.com/en-us/library/aa375680%28VS.85%29.aspx
                #
                # Under Windows XP and 2003, policies are stored as
                # <version>.policy files inside
                # %SystemRoot%\WinSxS\Policies\<name>
                # Under Vista and later, policies are stored as
                # <name>.manifest files inside %SystemRoot%\winsxs\Manifests
                redirected = False
                if os.path.isdir(pcfiles):
                    if not silent:
                        print(
                            "I: Searching for publisher configuration %s..."
                            % self.getpolicyid(True, language=language)
                        )
                    if sys.getwindowsversion() < (6,):
                        # Windows XP
                        policies = os.path.join(
                            pcfiles,
                            self.getpolicyid(True, language=language) + ".policy",
                        )
                    else:
                        # Vista or later
                        policies = os.path.join(
                            pcfiles,
                            self.getpolicyid(True, language=language) + ".manifest",
                        )
                    for manifestpth in glob(policies):
                        if not os.path.isfile(manifestpth):
                            if not silent:
                                print("W: Not a file", manifestpth)
                            continue
                        if not silent:
                            print("I: Found", manifestpth)
                        try:
                            policy = ManifestFromXMLFile(manifestpth)
                        except Exception as exc:
                            print("E: Could not parse file", manifestpth)
                            print("E:", str(exc))
                        else:
                            if not silent:
                                print(
                                    "I: Checking publisher policy for "
                                    "binding redirects"
                                )
                            for assembly in policy.dependentAssemblies:
                                if (
                                    not assembly.same_id(self, True)
                                    or assembly.optional
                                ):
                                    continue
                                for redirect in assembly.bindingRedirects:
                                    if not silent:
                                        old = "-".join(
                                            [
                                                ".".join([str(i) for i in part])
                                                for part in redirect[0]
                                            ]
                                        )
                                        new = ".".join([str(i) for i in redirect[1]])
                                        print(
                                            "I: Found redirect for version(s)",
                                            old,
                                            "->",
                                            new,
                                        )
                                    if (
                                        redirect[0][0] <= version <= redirect[0][-1]
                                        and version != redirect[1]
                                    ):
                                        if not silent:
                                            print(
                                                "I: Applying redirect",
                                                ".".join([str(i) for i in version]),
                                                "->",
                                                new,
                                            )
                                        version = redirect[1]
                                        redirected = True
                    if not redirected and not silent:
                        print("I: Publisher configuration not used")

            # Search for assemblies according to assembly searching sequence
            paths = []
            if os.path.isdir(manifests):
                # Add winsxs search paths
                paths.extend(
                    glob(
                        os.path.join(
                            manifests,
                            self.getid(language=language, version=version)
                            + "_*.manifest",
                        )
                    )
                )
            if self.filename:
                # Add private assembly search paths
                dirnm = os.path.dirname(self.filename)
                if language in (LANGUAGE_NEUTRAL_NT5, LANGUAGE_NEUTRAL_NT6):
                    for ext in (".dll", ".manifest"):
                        paths.extend(glob(os.path.join(dirnm, self.name + ext)))
                        paths.extend(
                            glob(os.path.join(dirnm, self.name, self.name + ext))
                        )
                else:
                    for ext in (".dll", ".manifest"):
                        paths.extend(
                            glob(os.path.join(dirnm, language, self.name + ext))
                        )
                    for ext in (".dll", ".manifest"):
                        paths.extend(
                            glob(
                                os.path.join(
                                    dirnm, language, self.name, self.name + ext
                                )
                            )
                        )
            if not silent:
                print(
                    "I: Searching for assembly %s..."
                    % self.getid(language=language, version=version)
                )
            for manifestpth in paths:
                if not os.path.isfile(manifestpth):
                    if not silent:
                        print("W: Not a file", manifestpth)
                    continue
                assemblynm = os.path.basename(os.path.splitext(manifestpth)[0])
                if not silent:
                    if manifestpth.endswith(".dll"):
                        print("I: Found manifest in", manifestpth)
                    else:
                        print("I: Found manifest", manifestpth)
                try:
                    if manifestpth.endswith(".dll"):
                        manifest = ManifestFromResFile(manifestpth, [1])
                    else:
                        manifest = ManifestFromXMLFile(manifestpth)
                except Exception as exc:
                    print("E: Could not parse manifest", manifestpth)
                    print("E:", exc)
                else:
                    if manifestpth.startswith(winsxs):
                        assemblydir = os.path.join(winsxs, assemblynm)
                        if not os.path.isdir(assemblydir):
                            if not silent:
                                print("W: No such dir", assemblydir)
                                print("W: Assembly incomplete")
                            return []
                    else:
                        assemblydir = os.path.dirname(manifestpth)
                    files.append(manifestpth)
                    for file_ in self.files or manifest.files:
                        fn = file_.find(assemblydir)
                        if fn:
                            files.append(fn)
                        else:
                            # If any of our files does not exist,
                            # the assembly is incomplete
                            if not silent:
                                print("W: Assembly incomplete")
                            return []
                return files

        print("W: Assembly not found")
        return []

    def getid(self, language=None, version=None):
        """Return an identification string which uniquely names a manifest.

        This string is a combination of the manifest's processorArchitecture,
        name, publicKeyToken, version and language.

        Arguments:
        version (tuple or list of integers) - If version is given, use it
                                              instead of the manifest's
                                              version.

        """
        if not self.name:
            if not silent:
                print("W: Assembly metadata incomplete")
            return ""
        id = []
        if self.processorArchitecture:
            id.append(self.processorArchitecture)
        id.append(self.name)
        if self.publicKeyToken:
            id.append(self.publicKeyToken)
        if version or self.version:
            id.append(".".join([str(i) for i in version or self.version]))
        if not language:
            language = self.getlanguage()
        if language:
            id.append(language)
        return "_".join(id)

    def getlanguage(self, language=None, windowsversion=None):
        """Get and return the manifest's language as string.

        Can be either language-culture e.g. 'en-us' or a string indicating
        language neutrality, e.g. 'x-ww' on Windows XP or 'none' on Vista
        and later.

        """
        if not language:
            language = self.language
        if language in (None, "", "*", "neutral"):
            return (LANGUAGE_NEUTRAL_NT5, LANGUAGE_NEUTRAL_NT6)[
                (windowsversion or sys.getwindowsversion()) >= (6,)
            ]
        return language

    def getpolicyid(self, fuzzy=True, language=None, windowsversion=None):
        """Return an identification string which can be used to find a policy.

        This string is a combination of the manifest's processorArchitecture,
        major and minor version, name, publicKeyToken and language.

        Arguments:
        fuzzy (boolean)             - If False, insert the full version in
                                      the id string. Default is True (omit).
        windowsversion              - If not specified (or None), default to
        (tuple or list of integers)   sys.getwindowsversion().

        """
        if not self.name:
            if not silent:
                print("W: Assembly metadata incomplete")
            return ""
        id = []
        if self.processorArchitecture:
            id.append(self.processorArchitecture)
        name = []
        name.append("policy")
        if self.version:
            name.append(str(self.version[0]))
            name.append(str(self.version[1]))
        name.append(self.name)
        id.append(".".join(name))
        if self.publicKeyToken:
            id.append(self.publicKeyToken)
        if self.version and (windowsversion or sys.getwindowsversion()) >= (6,):
            # Vista and later
            if fuzzy:
                id.append("*")
            else:
                id.append(".".join([str(i) for i in self.version]))
        if not language:
            language = self.getlanguage(windowsversion=windowsversion)
        if language:
            id.append(language)
        id.append("*")
        id = "_".join(id)
        if self.version and (windowsversion or sys.getwindowsversion()) < (6,):
            # Windows XP
            if fuzzy:
                id = os.path.join(id, "*")
            else:
                id = os.path.join(id, ".".join([str(i) for i in self.version]))
        return id

    def load_dom(self, domtree, initialize=True):
        """Load manifest from DOM tree.

        If initialize is True (default), reset existing attributes first.

        """
        if domtree.nodeType == Node.DOCUMENT_NODE:
            rootElement = domtree.documentElement
        elif domtree.nodeType == Node.ELEMENT_NODE:
            rootElement = domtree
        else:
            raise InvalidManifestError(
                "Invalid root element node type %s - has to be one of "
                "(DOCUMENT_NODE, ELEMENT_NODE)" % domtree
            )
        allowed_names = (
            "assembly",
            "assemblyBinding",
            "configuration",
            "dependentAssembly",
        )
        if rootElement.tagName not in allowed_names:
            raise InvalidManifestError(
                "Invalid root element <%s> - has to be one of <%s>"
                % (rootElement.tagName, ">, <".join(allowed_names))
            )
        # print "I: loading manifest metadata from element <%s>" % \
        # rootElement.tagName
        if rootElement.tagName == "configuration":
            for windows in rootElement.getCEByTN("windows"):
                for assemblyBinding in windows.getCEByTN("assemblyBinding"):
                    self.load_dom(assemblyBinding, initialize)
        else:
            if initialize:
                self.__init__()
            self.manifestType = rootElement.tagName
            self.manifestVersion = [
                int(i)
                for i in (rootElement.getA("manifestVersion") or "1.0").split(".")
            ]
            self.noInheritable = bool(rootElement.getFCEByTN("noInheritable"))
            self.noInherit = bool(rootElement.getFCEByTN("noInherit"))
            for assemblyIdentity in rootElement.getCEByTN("assemblyIdentity"):
                self.type = assemblyIdentity.getA("type") or None
                self.name = assemblyIdentity.getA("name") or None
                self.language = assemblyIdentity.getA("language") or None
                self.processorArchitecture = (
                    assemblyIdentity.getA("processorArchitecture") or None
                )
                version = assemblyIdentity.getA("version")
                if version:
                    self.version = [int(i) for i in version.split(".")]
                self.publicKeyToken = assemblyIdentity.getA("publicKeyToken") or None
            for publisherPolicy in rootElement.getCEByTN("publisherPolicy"):
                self.applyPublisherPolicy = (
                    publisherPolicy.getA("apply") or ""
                ).lower() == "yes"
            for description in rootElement.getCEByTN("description"):
                if description.firstChild:
                    self.description = description.firstChild.wholeText
            for trustInfo in rootElement.getCEByTN("trustInfo"):
                for security in trustInfo.getCEByTN("security"):
                    for requestedPrivileges in security.getCEByTN(
                        "requestedPrivileges"
                    ):
                        for requestedExecutionLevel in requestedPrivileges.getCEByTN(
                            "requestedExecutionLevel"
                        ):
                            self.requestedExecutionLevel = requestedExecutionLevel.getA(
                                "level"
                            )
                            self.uiAccess = (
                                requestedExecutionLevel.getA("uiAccess") or ""
                            ).lower() == "true"
            if rootElement.tagName == "assemblyBinding":
                dependencies = [rootElement]
            else:
                dependencies = rootElement.getCEByTN("dependency")
            for dependency in dependencies:
                for dependentAssembly in dependency.getCEByTN("dependentAssembly"):
                    manifest = ManifestFromDOM(dependentAssembly)
                    if not manifest.name:
                        # invalid, skip
                        continue
                    manifest.optional = (
                        dependency.getA("optional") or ""
                    ).lower() == "yes"
                    self.dependentAssemblies.append(manifest)
                    if self.filename:
                        # Enable search for private assembly by assigning bogus
                        # filename (only the directory has to be correct)
                        self.dependentAssemblies[-1].filename = ":".join(
                            (self.filename, manifest.name)
                        )
            for bindingRedirect in rootElement.getCEByTN("bindingRedirect"):
                oldVersion = [
                    [int(i) for i in part.split(".")]
                    for part in bindingRedirect.getA("oldVersion").split("-")
                ]
                newVersion = [
                    int(i) for i in bindingRedirect.getA("newVersion").split(".")
                ]
                self.bindingRedirects.append((oldVersion, newVersion))
            for file_ in rootElement.getCEByTN("file"):
                self.add_file(
                    name=file_.getA("name"),
                    hashalg=file_.getA("hashalg"),
                    hash=file_.getA("hash"),
                )

    def parse(self, filename_or_file, initialize=True):
        """Load manifest from file or file object"""
        if isinstance(filename_or_file, str):
            filename = filename_or_file
        else:
            filename = filename_or_file.name
        try:
            domtree = minidom.parse(filename_or_file)
        except xml.parsers.expat.ExpatError as e:
            args = [e.args[0]]
            if isinstance(filename, str):
                filename = filename.encode(sys.getdefaultencoding(), "replace")
            args.insert(0, '\n  File "%s"\n   ' % filename)
            raise ManifestXMLParseError(" ".join([str(arg) for arg in args]))
        if initialize:
            self.__init__()
        self.filename = filename
        self.load_dom(domtree, False)

    def parse_string(self, xmlstr, initialize=True):
        """Load manifest from XML string"""
        try:
            domtree = minidom.parseString(xmlstr)
        except xml.parsers.expat.ExpatError as e:
            raise ManifestXMLParseError(e)
        self.load_dom(domtree, initialize)

    def same_id(self, manifest, skip_version_check=False):
        """Return a bool indicating if another manifest has the same identitiy.

        This is done by comparing language, name, processorArchitecture,
        publicKeyToken, type and version.

        """
        if skip_version_check:
            version_check = True
        else:
            version_check = self.version == manifest.version
        return (
            self.language == manifest.language
            and self.name == manifest.name
            and self.processorArchitecture == manifest.processorArchitecture
            and self.publicKeyToken == manifest.publicKeyToken
            and self.type == manifest.type
            and version_check
        )

    def todom(self):
        """Return the manifest as DOM tree"""
        doc = Document()
        docE = doc.cE(self.manifestType)
        if self.manifestType == "assemblyBinding":
            cfg = doc.cE("configuration")
            win = doc.cE("windows")
            win.aChild(docE)
            cfg.aChild(win)
            doc.aChild(cfg)
        else:
            doc.aChild(docE)
        if self.manifestType != "dependentAssembly":
            docE.setA("xmlns", "urn:schemas-microsoft-com:asm.v1")
            if self.manifestType != "assemblyBinding":
                docE.setA(
                    "manifestVersion", ".".join([str(i) for i in self.manifestVersion])
                )
        if self.noInheritable:
            docE.aChild(doc.cE("noInheritable"))
        if self.noInherit:
            docE.aChild(doc.cE("noInherit"))
        aId = doc.cE("assemblyIdentity")
        if self.type:
            aId.setAttribute("type", self.type)
        if self.name:
            aId.setAttribute("name", self.name)
        if self.language:
            aId.setAttribute("language", self.language)
        if self.processorArchitecture:
            aId.setAttribute("processorArchitecture", self.processorArchitecture)
        if self.version:
            aId.setAttribute("version", ".".join([str(i) for i in self.version]))
        if self.publicKeyToken:
            aId.setAttribute("publicKeyToken", self.publicKeyToken)
        if aId.hasAttributes():
            docE.aChild(aId)
        else:
            aId.unlink()
        if self.applyPublisherPolicy is not None:
            ppE = doc.cE("publisherPolicy")
            if self.applyPublisherPolicy:
                ppE.setA("apply", "yes")
            else:
                ppE.setA("apply", "no")
            docE.aChild(ppE)
        if self.description:
            descE = doc.cE("description")
            descE.aChild(doc.cT(self.description))
            docE.aChild(descE)
        if self.requestedExecutionLevel in (
            "asInvoker",
            "highestAvailable",
            "requireAdministrator",
        ):
            tE = doc.cE("trustInfo")
            tE.setA("xmlns", "urn:schemas-microsoft-com:asm.v3")
            sE = doc.cE("security")
            rpE = doc.cE("requestedPrivileges")
            relE = doc.cE("requestedExecutionLevel")
            relE.setA("level", self.requestedExecutionLevel)
            if self.uiAccess:
                relE.setA("uiAccess", "true")
            else:
                relE.setA("uiAccess", "false")
            rpE.aChild(relE)
            sE.aChild(rpE)
            tE.aChild(sE)
            docE.aChild(tE)
        if self.dependentAssemblies:
            for assembly in self.dependentAssemblies:
                if self.manifestType != "assemblyBinding":
                    dE = doc.cE("dependency")
                    if assembly.optional:
                        dE.setAttribute("optional", "yes")
                daE = doc.cE("dependentAssembly")
                adom = assembly.todom()
                for child in adom.documentElement.childNodes:
                    daE.aChild(child.cloneNode(False))
                adom.unlink()
                if self.manifestType != "assemblyBinding":
                    dE.aChild(daE)
                    docE.aChild(dE)
                else:
                    docE.aChild(daE)
        if self.bindingRedirects:
            for bindingRedirect in self.bindingRedirects:
                brE = doc.cE("bindingRedirect")
                brE.setAttribute(
                    "oldVersion",
                    "-".join(
                        [
                            ".".join([str(i) for i in part])
                            for part in bindingRedirect[0]
                        ]
                    ),
                )
                brE.setAttribute(
                    "newVersion", ".".join([str(i) for i in bindingRedirect[1]])
                )
                docE.aChild(brE)
        if self.files:
            for file_ in self.files:
                fE = doc.cE("file")
                for attr in ("name", "hashalg", "hash"):
                    val = getattr(file_, attr)
                    if val:
                        fE.setA(attr, val)
                docE.aChild(fE)
        return doc

    def toprettyxml(self, indent="  ", newl=os.linesep, encoding="UTF-8"):
        """Return the manifest as pretty-printed XML"""
        domtree = self.todom()
        # WARNING: The XML declaration has to follow the order
        # version-encoding-standalone (standalone being optional), otherwise
        # if it is embedded in an exe the exe will fail to launch!
        # ('application configuration incorrect')
        if sys.version_info >= (2, 3):
            xmlstr = domtree.toprettyxml(indent, newl, encoding)
        else:
            xmlstr = domtree.toprettyxml(indent, newl)
        xmlstr = xmlstr.strip(os.linesep.encode('utf-8')).replace(
            ('<?xml version="1.0" encoding="%s"?>' % encoding).encode('utf-8'),
            ('<?xml version="1.0" encoding="%s" standalone="yes"?>' % encoding).encode('utf-8'),
        )
        domtree.unlink()
        return xmlstr

    def toxml(self, encoding="UTF-8"):
        """Return the manifest as XML"""
        domtree = self.todom()
        # WARNING: The XML declaration has to follow the order
        # version-encoding-standalone (standalone being optional), otherwise
        # if it is embedded in an exe the exe will fail to launch!
        # ('application configuration incorrect')
        xmlstr = domtree.toxml(encoding).replace(
            '<?xml version="1.0" encoding="%s"?>' % encoding,
            '<?xml version="1.0" encoding="%s" standalone="yes"?>' % encoding,
        )
        domtree.unlink()
        return xmlstr

    def update_resources(self, dstpath, names=None, languages=None):
        """Update or add manifest resource in dll/exe file dstpath"""
        UpdateManifestResourcesFromXML(dstpath, self.toprettyxml(), names, languages)

    def writeprettyxml(
        self, filename_or_file=None, indent="  ", newl=os.linesep, encoding="UTF-8"
    ):
        """Write the manifest as XML to a file or file object"""
        if not filename_or_file:
            filename_or_file = self.filename
        if isinstance(filename_or_file, str):
            filename_or_file = open(filename_or_file, "wb")
        xmlstr = self.toprettyxml(indent, newl, encoding)
        filename_or_file.write(xmlstr)
        filename_or_file.close()

    def writexml(
        self, filename_or_file=None, indent="  ", newl=os.linesep, encoding="UTF-8"
    ):
        """Write the manifest as XML to a file or file object"""
        if not filename_or_file:
            filename_or_file = self.filename
        if isinstance(filename_or_file, str):
            filename_or_file = open(filename_or_file, "wb")
        xmlstr = self.toxml(indent, newl, encoding)
        filename_or_file.write(xmlstr)
        filename_or_file.close()


def ManifestFromResFile(filename, names=None, languages=None):
    """Create and return manifest instance from resource in dll/exe file"""
    res = GetManifestResources(filename, names, languages)
    pth = []
    if res and res[RT_MANIFEST]:
        while isinstance(res, dict) and res.keys():
            key = res.keys()[0]
            pth.append(str(key))
            res = res[key]
    if isinstance(res, dict):
        raise InvalidManifestError(
            "No matching manifest resource found in '%s'" % filename
        )
    manifest = Manifest()
    manifest.filename = ":".join([filename] + pth)
    manifest.parse_string(res, False)
    return manifest


def ManifestFromDOM(domtree):
    """Create and return manifest instance from DOM tree"""
    manifest = Manifest()
    manifest.load_dom(domtree)
    return manifest


def ManifestFromXML(xmlstr):
    """Create and return manifest instance from XML"""
    manifest = Manifest()
    manifest.parse_string(xmlstr)
    return manifest


def ManifestFromXMLFile(filename_or_file):
    """Create and return manifest instance from file"""
    manifest = Manifest()
    manifest.parse(filename_or_file)
    return manifest


def GetManifestResources(filename, names=None, languages=None):
    """Get manifest resources from file"""
    return winresource.GetResources(filename, [RT_MANIFEST], names, languages)


def UpdateManifestResourcesFromXML(dstpath, xmlstr, names=None, languages=None):
    """Update or add manifest XML as resource in dstpath"""
    if not silent:
        print("I: Updating manifest in", dstpath)
    if dstpath.lower().endswith(".exe"):
        name = 1
    else:
        name = 2
    winresource.UpdateResources(
        dstpath, xmlstr, RT_MANIFEST, names or [name], languages or [0, "*"]
    )


def UpdateManifestResourcesFromXMLFile(dstpath, srcpath, names=None, languages=None):
    """Update or add manifest XML from srcpath as resource in dstpath"""
    if not silent:
        print("I: Updating manifest from", srcpath, "in", dstpath)
    if dstpath.lower().endswith(".exe"):
        name = 1
    else:
        name = 2
    winresource.UpdateResourcesFromDataFile(
        dstpath, srcpath, RT_MANIFEST, names or [name], languages or [0, "*"]
    )


if __name__ == "__main__":
    dstpath = sys.argv[1]
    srcpath = sys.argv[2]
    UpdateManifestResourcesFromXMLFile(dstpath, srcpath)



# -*- coding: utf-8 -*-

import demjson_compat

from DisplayCAL.lazydict import LazyDict


class JSONDict(LazyDict):
    """JSON lazy dictionary"""

    def parse(self, fileobj):
        dict.update(self, demjson_compat.decode(fileobj.read()))



# -*- coding: utf-8 -*-
import os
from subprocess import Popen

from DisplayCAL import worker_base
from DisplayCAL import config
from DisplayCAL.dev.mocks import check_call

from tests.data.argyll_sp_data import SUBPROCESS_COM

# todo: deactivated test temporarily
# def test_xicclu_is_working_properly(data_files):
#     """testing if ``DisplayCAL.worker_base.Xicclu`` is working properly"""
#     from DisplayCAL import ICCProfile
#     from DisplayCAL.worker_base import Xicclu
#
#     profile = ICCProfile.ICCProfile(profile=data_files["default.icc"].absolute())
#     xicclu = Xicclu(profile, "r", "a", pcs="X", scale=100)
#     assert xicclu() is not None


def test_get_argyll_util(argyll):
    """Test worker_base.get_argyll_util() function."""
    config.initcfg()
    result = worker_base.get_argyll_util("ccxxmake")
    expected_result = os.path.join(config.getcfg("argyll.dir"), "ccxxmake")
    assert result == expected_result


def test_get_argyll_version_string_1(argyll):
    """Test worker_base.get_argyll_version_string() function."""
    config.initcfg()
    with check_call(Popen, "communicate", SUBPROCESS_COM):
        result = worker_base.get_argyll_version_string("ccxxmake")
    expected_result = "2.3.0"
    assert result == expected_result


def test_printcmdline_1():
    """Test worker_base.printcmdline() function for issue #73"""
    # Command line:
    cmd = "/home/eoyilmaz/.local/bin/Argyll_V2.3.0/bin/colprof"
    args = [
        "-v",
        "-qh",
        "-ax",
        "-bn",
        "-C",
        b"No copyright. Created with DisplayCAL 3.8.9.3 and Argyll CMS 2.3.0",
        "-A",
        "Dell, Inc.",
        "-D",
        "UP2516D_#1_2022-04-01_00-26_2.2_F-S_XYZLUT+MTX",
        "/tmp/DisplayCAL-i91d9z8_/UP2516D_#1_2022-04-01_00-26_2.2_F-S_XYZLUT+MTX",
    ]
    # fn = "<bound method WorkerBase.log of <DisplayCAL.worker.Worker object at 0x7f7b941bb6a0>>"
    cwd = "/tmp/DisplayCAL-i91d9z8_"
    worker_base.printcmdline(cmd=cmd, args=args, cwd=cwd)



# -*- coding: utf-8 -*-
"""
Parse commandline options.

Note that none of these are advertised, as they solely exist for testing and development
purposes.
"""

import sys

# Use only ascii? (DON'T)
ascii = "--ascii" in sys.argv[1:]

# Debug level (default: 0 = off). >= 1 prints debug messages
if "-d2" in sys.argv[1:] or "--debug=2" in sys.argv[1:]:
    debug = 2
elif (
    "-d1" in sys.argv[1:]
    or "--debug=1" in sys.argv[1:]
    or "-d" in sys.argv[1:]
    or "--debug" in sys.argv[1:]
):
    debug = 1
else:
    debug = 0

# Debug localization
debug_localization = "-dl" in sys.argv[1:] or "--debug-localization" in sys.argv[1:]

# Use alternate patch preview in the testchart editor?
tc_use_alternate_preview = (
    "-ap" in sys.argv[1:] or "--alternate-preview" in sys.argv[1:]
)

# Test some features even if they are not available normally
test = "-t" in sys.argv[1:] or "--test" in sys.argv[1:]

eecolor65 = "--ee65" in sys.argv[1:]

# Test sensor calibration
test_require_sensor_cal = (
    "-s" in sys.argv[1:] or "--test_require_sensor_cal" in sys.argv[1:]
)

# Test update functionality
test_update = "-tu" in sys.argv[1:] or "--test-update" in sys.argv[1:]

# Test SSL connection using badssl.com
test_badssl = False
for arg in sys.argv[1:]:
    if arg.startswith("--test-badssl="):
        test_badssl = arg.split("=", 1)[-1]

# Always fail download
always_fail_download = "--always-fail-download" in sys.argv[1:]

# HDR input profile generation: Test input curve clipping
test_input_curve_clipping = "--test-input-curve-clipping" in sys.argv[1:]

# Enable experimental features
experimental = "-x" in sys.argv[1:] or "--experimental" in sys.argv[1:]

# Verbosity level (default: 1). >= 1 prints some status information
if "-v4" in sys.argv[1:] or "--verbose=4" in sys.argv[1:]:
    verbose = 4
elif "-v3" in sys.argv[1:] or "--verbose=3" in sys.argv[1:]:
    verbose = 3
elif "-v2" in sys.argv[1:] or "--verbose=2" in sys.argv[1:]:
    verbose = 2
elif "-v0" in sys.argv[1:] or "--verbose=0" in sys.argv[1:]:
    verbose = 0  # Off
else:
    verbose = 1

# Use Colord GObject introspection interface (otherwise, use D-Bus)
use_colord_gi = "--use-colord-gi" in sys.argv[1:]

# Skip initial instrument/port detection on startup
force_skip_initial_instrument_detection = (
    "--force-skip-initial-instrument-detection" in sys.argv[1:]
)



# -*- coding: utf-8 -*-
from time import sleep
import os
import sys
import threading
import types

from DisplayCAL.colormath import specialpow
from DisplayCAL.wxfixes import wx, GenButton, PlateButton, get_dialogs

import wx.grid
from DisplayCAL.lib.agw.gradientbutton import GradientButton
from DisplayCAL import floatspin


def AdjustMinMax(self, minvalue=0.0, maxvalue=1.0):
    """Adjust min/max"""
    buffer = self.GetDataBuffer()
    for i, byte in enumerate(buffer):
        buffer[i] = min(int(round(minvalue * 255 + byte * (maxvalue - minvalue))), 255)


wx.Image.AdjustMinMax = AdjustMinMax


def Blend(self, bitmap, x, y):
    """Blend the given bitmap over the specified position in this bitmap."""
    dc = wx.MemoryDC(self)
    dc.DrawBitmap(bitmap, x, y)


wx.Bitmap.Blend = Blend


def Invert(self):
    """Invert image colors"""
    databuffer = self.GetDataBuffer()
    for i, byte in enumerate(databuffer):
        databuffer[i] = 255 - byte


wx.Image.Invert = Invert


def GammaCorrect(self, from_gamma=-2.4, to_gamma=1.8):
    """Gamma correct"""
    buffer = self.GetDataBuffer()
    for i, byte in enumerate(buffer):
        buffer[i] = int(
            round(specialpow(byte / 255.0, from_gamma) ** (1.0 / to_gamma) * 255)
        )


wx.Image.GammaCorrect = GammaCorrect


def IsBW(self):
    """Check if image is grayscale in the most effective way possible.

    Note that this is a costly operation even though it returns as quickly as
    possible for non-grayscale images (i.e. when it encounters the first
    non-equal RGB triplet).

    """
    triplet = set()
    for i, byte in enumerate(self.GetDataBuffer()):
        triplet.add(byte)
        if i % 3 == 2:
            if len(triplet) != 1:
                return False
            triplet = set()
    return True


wx.Image.IsBW = IsBW


def GetRealClientArea(self):
    """Return the real (non-overlapping) client area of a display"""
    # need to fix overlapping ClientArea on some Linux multi-display setups
    # the client area must be always smaller than the geometry
    clientarea = list(self.ClientArea)
    if self.Geometry[0] > clientarea[0]:
        clientarea[0] = self.Geometry[0]
    if self.Geometry[1] > clientarea[1]:
        clientarea[1] = self.Geometry[1]
    if self.Geometry[2] < clientarea[2]:
        clientarea[2] = self.Geometry[2]
    if self.Geometry[3] < clientarea[3]:
        clientarea[3] = self.Geometry[3]
    return wx.Rect(*clientarea)


wx.Display.GetRealClientArea = GetRealClientArea


def GetAllChildren(self, skip=None):
    """Get children of window and its subwindows"""
    if not isinstance(skip, (list, tuple)):
        skip = [skip]
    children = [child for child in self.GetChildren() if child not in skip]
    allchildren = []
    for child in children:
        allchildren.append(child)
        if hasattr(child, "GetAllChildren") and callable(child.GetAllChildren):
            allchildren += child.GetAllChildren(skip)
    return allchildren


wx.Window.GetAllChildren = GetAllChildren


def GetDisplay(self):
    """Return the display the window is shown on"""
    display_no = wx.Display.GetFromWindow(self)
    if display_no < 0:  # window outside visible area
        display_no = 0
    return wx.Display(display_no)


wx.Window.GetDisplay = GetDisplay


def SetMaxFontSize(self, pointsize=11):
    font = self.GetFont()
    if font.GetPointSize() > pointsize:
        font.SetPointSize(pointsize)
        self.SetFont(font)


wx.Window.SetMaxFontSize = SetMaxFontSize


def RealCenterOnScreen(self, dir=wx.BOTH):
    """Center the window on the screen it is on, unlike CenterOnScreen which
    always centers on 1st screen.

    """
    x, y = self.Position
    left, top, w, h = self.GetDisplay().ClientArea
    if dir & wx.HORIZONTAL:
        x = left + w / 2 - self.Size[0] / 2
    if dir & wx.VERTICAL:
        y = top + h / 2 - self.Size[1] / 2
    self.Position = x, y


wx.TopLevelWindow.RealCenterOnScreen = RealCenterOnScreen


def SetSaneGeometry(self, x=None, y=None, w=None, h=None):
    """Set a 'sane' window position and/or size (within visible screen area)."""
    if None not in (x, y):
        # First, move to coordinates given
        self.SetPosition((x, y))
    # Returns the first display's client area if the window
    # is completely outside the client area of all displays
    display_client_rect = self.GetDisplay().ClientArea
    if sys.platform not in ("darwin", "win32"):  # Linux
        if os.getenv("XDG_SESSION_TYPE") == "wayland":
            # Client-side decorations
            safety_margin = 0
        else:
            # Assume server-side decorations
            safety_margin = 40
    else:
        safety_margin = 20
    if None not in (w, h):
        # Set given size, but resize if needed to fit inside client area
        if hasattr(self, "MinClientSize"):
            min_w, min_h = self.MinClientSize
        else:
            min_w, min_h = self.WindowToClientSize(self.MinSize)
        border_lr = self.Size[0] - self.ClientSize[0]
        border_tb = self.Size[1] - self.ClientSize[1]
        self.ClientSize = (
            min(display_client_rect[2] - border_lr, max(w, min_w)),
            min(display_client_rect[3] - border_tb - safety_margin, max(h, min_h)),
        )
    if None not in (x, y):
        if not display_client_rect.ContainsXY(
            x, y
        ) or not display_client_rect.ContainsRect((x, y, self.Size[0], self.Size[1])):
            # If outside client area, move into client area
            xy = [x, y]
            for i, pos in enumerate([xy, (x + self.Size[0], y + self.Size[1])]):
                for j in range(2):
                    if (
                        pos[j] > display_client_rect[j] + display_client_rect[2 + j]
                        or pos[j] < display_client_rect[j]
                    ):
                        if i:
                            xy[j] = (
                                display_client_rect[j]
                                + display_client_rect[2 + j]
                                - self.Size[j]
                            )
                        else:
                            xy[j] = display_client_rect[j]
            self.SetPosition(tuple(xy))


wx.Window.SetSaneGeometry = SetSaneGeometry


def GridGetSelectedRowsFromSelection(self):
    """Return the number of fully selected rows.

    Unlike GetSelectedRows, include rows that have been selected
    by chosing individual cells.

    """
    numcols = self.GetNumberCols()
    rows = []
    i = -1
    for cell in self.GetSelection():
        row, col = cell
        if row > i:
            i = row
            rownumcols = 0
        rownumcols += 1
        if rownumcols == numcols:
            rows.append(row)
    return rows


wx.grid.Grid.GetSelectedRowsFromSelection = GridGetSelectedRowsFromSelection


def GridGetSelectionRows(self):
    """Return the selected rows, even if not all cells in a row are selected."""
    rows = []
    i = -1
    for row, _col in self.GetSelection():
        if row > i:
            i = row
            rows.append(row)
    return rows


wx.grid.Grid.GetSelectionRows = GridGetSelectionRows


def IsSizer(self):
    """Check if the window is a sizer"""
    return isinstance(self, wx.Sizer)


wx.Window.IsSizer = IsSizer


def gamma_encode(R, G, B, alpha=wx.ALPHA_OPAQUE):
    """(Re-)Encode R'G'B' colors with specific platform gamma.

    R, G, B = color components in range 0..255

    Note this only has effect under wxMac which assumes a decoding gamma of 1.8

    """
    if sys.platform == "darwin":
        # Adjust for wxMac assuming gamma 1.8 instead of sRGB
        # Decode R'G'B' -> linear light using sRGB transfer function, then
        # re-encode to gamma = 1.0 / 1.8 so that when decoded with gamma = 1.8
        # we get the correct sRGB color
        RGBa = [
            int(round(specialpow(v / 255.0, -2.4) ** (1.0 / 1.8) * 255))
            for v in (R, G, B)
        ]
        RGBa.append(alpha)
        return RGBa
    else:
        return [R, G, B, alpha]


def get_platform_window_decoration_size():
    if sys.platform in ("darwin", "win32"):
        # Size includes windows decoration
        if sys.platform == "win32":
            border = 8  # Windows 7
            titlebar = 30  # Windows 7
        else:
            border = 0  # Mac OS X 10.7 Lion
            titlebar = 22  # Mac OS X 10.7 Lion
    else:
        # Linux. Size does not include window decoration
        border = 0
        titlebar = 0
    return border, titlebar


def draw_granger_rainbow(dc, x=0, y=0, width=1920, height=1080):
    """Draw a granger rainbow to a DC"""
    if not isinstance(dc, wx.GCDC):
        raise NotImplementedError("%s lacks alpha transparency support" % dc.__class__)

    # Widths
    column_width = int(162.0 / 1920.0 * width)
    rainbow_width = width - column_width * 2
    strip_width = int(rainbow_width / 7.0)
    rainbow_width = strip_width * 7
    column_width = (width - rainbow_width) / 2

    # Gray columns left/right
    dc.GradientFillLinear(
        wx.Rect(x, y, width, height),
        wx.Colour(0, 0, 0),
        wx.Colour(255, 255, 255),
        wx.UP,
    )

    # Granger rainbow
    rainbow = [
        (255, 0, 255),
        (0, 0, 255),
        (0, 255, 255),
        (0, 255, 0),
        (255, 255, 0),
        (255, 0, 0),
        (255, 0, 255),
    ]
    start = rainbow[-2]
    for i, end in enumerate(rainbow):
        dc.GradientFillLinear(
            wx.Rect(x + column_width + strip_width * i, y, strip_width, height),
            wx.Colour(*start),
            wx.Colour(*end),
            wx.RIGHT,
        )
        start = end

    # White-to-black gradient with transparency for shading
    # Top half - white to transparent
    dc.GradientFillLinear(
        wx.Rect(x + column_width, y, rainbow_width, height / 2),
        wx.Colour(0, 0, 0, 0),
        wx.Colour(255, 255, 255, 255),
        wx.UP,
    )
    # Bottom half - transparent to black
    dc.GradientFillLinear(
        wx.Rect(x + column_width, y + height / 2, rainbow_width, height / 2),
        wx.Colour(0, 0, 0, 255),
        wx.Colour(255, 255, 255, 0),
        wx.UP,
    )


def create_granger_rainbow_bitmap(width=1920, height=1080, filename=None):
    """Create a granger rainbow bitmap"""
    # Main bitmap
    bmp = wx.EmptyBitmap(width, height)
    mdc = wx.MemoryDC(bmp)
    dc = wx.GCDC(mdc)

    draw_granger_rainbow(dc, 0, 0, width, height)

    mdc.SelectObject(wx.NullBitmap)

    if filename:
        name, ext = os.path.splitext(filename)
        ext = ext.lower()
        bmp_type = {
            ".bmp": wx.BITMAP_TYPE_BMP,
            ".jpe": wx.BITMAP_TYPE_JPEG,
            ".jpg": wx.BITMAP_TYPE_JPEG,
            ".jpeg": wx.BITMAP_TYPE_JPEG,
            ".jfif": wx.BITMAP_TYPE_JPEG,
            ".pcx": wx.BITMAP_TYPE_PCX,
            ".png": wx.BITMAP_TYPE_PNG,
            ".tga": wx.BITMAP_TYPE_TGA,
            ".tif": wx.BITMAP_TYPE_TIFF,
            ".tiff": wx.BITMAP_TYPE_TIFF,
            ".xpm": wx.BITMAP_TYPE_XPM,
        }.get(ext, wx.BITMAP_TYPE_PNG)
        bmp.SaveFile(filename, bmp_type)
    else:
        return bmp


def get_parent_frame(window):
    """Get parent frame (if any)"""
    parent = window.Parent
    while parent:
        if isinstance(parent, wx.Frame):
            return parent
        parent = parent.Parent


class CustomEvent(wx.PyEvent):
    def __init__(self, typeId, object, window=None):
        wx.PyEvent.__init__(self, object.GetId(), typeId)
        self.EventObject = object
        self.Window = window

    def GetWindow(self):
        return self.Window


_global_timer_lock = threading.Lock()


wxEVT_BETTERTIMER = wx.NewEventType()
EVT_BETTERTIMER = wx.PyEventBinder(wxEVT_BETTERTIMER, 1)


class BetterTimerEvent(wx.PyCommandEvent):
    def __init__(self, id=wx.ID_ANY, ms=0):
        wx.PyCommandEvent.__init__(self, wxEVT_BETTERTIMER, id)
        self._ms = ms

    def GetInterval(self):
        return self._ms

    Interval = property(GetInterval)


class BetterTimer(wx.Timer):
    """A wx.Timer replacement.

    Doing GUI updates using regular timers can be incredibly segfaulty under
    wxPython Phoenix when several timers run concurrently.

    This approach uses a global lock to work around the issue.

    """

    def __init__(self, owner=None, timerid=wx.ID_ANY):
        wx.Timer.__init__(self, None, timerid)
        self._owner = owner

    def Notify(self):
        if self._owner and _global_timer_lock.acquire(False):
            try:
                wx.PostEvent(self._owner, BetterTimerEvent(self.Id, self.Interval))
            finally:
                _global_timer_lock.release()


class BetterCallLater(wx.CallLater):
    def __init__(self, millis, callableObj, *args, **kwargs):
        wx.CallLater.__init__(self, millis, callableObj, *args, **kwargs)

    def Notify(self):
        if _global_timer_lock.acquire(True):
            try:
                wx.CallLater.Notify(self)
            finally:
                _global_timer_lock.release()


class ThreadedTimer(object):
    """A wx.Timer replacement that uses threads instead of actual timers
    which are a limited resource"""

    def __init__(self, owner=None, timerid=wx.ID_ANY):
        self._owner = owner
        if timerid < 0:
            timerid = wx.Window.NewControlId()
        self._id = timerid
        self._ms = 0
        self._oneshot = False
        self._keep_running = False
        self._thread = None

    def _notify(self):
        if _global_timer_lock.acquire(self._oneshot):
            try:
                self.Notify()
            finally:
                _global_timer_lock.release()

    def _timer(self):
        self._keep_running = True
        while self._keep_running:
            sleep(self._ms / 1000.0)
            if self._keep_running:
                wx.CallAfter(self._notify)
                if self._oneshot:
                    self._keep_running = False

    def Destroy(self):
        if hasattr(wx.Window, "UnreserveControlId") and self.Id < 0:
            try:
                wx.Window.UnreserveControlId(self.Id)
            except wx.wxAssertionError:
                pass

    def GetId(self):
        return self._id

    def GetInterval(self):
        return self._ms

    def GetOwner(self):
        return self._owner

    def SetOwner(self, owner):
        self._owner = owner

    Id = property(GetId)
    Interval = property(GetInterval)
    Owner = property(GetOwner, SetOwner)

    def IsOneShot(self):
        return self._oneshot

    def IsRunning(self):
        return self._keep_running

    def Notify(self):
        if self._owner:
            self._owner.ProcessEvent(BetterTimerEvent(self._id, self._ms))

    def Start(self, milliseconds=-1, oneShot=False):
        if self._thread and self._thread.is_alive():
            self._keep_running = False
            self._thread.join()
        if milliseconds > -1:
            self._ms = milliseconds
        self._oneshot = oneShot
        self._thread = threading.Thread(target=self._timer, name="ThreadedTimer")
        self._thread.start()

    def Stop(self):
        self._keep_running = False


class ThreadedCallLater(ThreadedTimer):
    def __init__(self, millis, callableObj, *args, **kwargs):
        ThreadedTimer.__init__(self)
        self._oneshot = True
        self._callable = callableObj
        self._has_run = False
        self._result = None
        self.SetArgs(*args, **kwargs)
        self.Start(millis)

    def GetResult(self):
        return self._result

    Result = property(GetResult)

    def HasRun(self):
        return self._has_run

    def Notify(self):
        self._result = self._callable(*self._args, **self._kwargs)
        self._has_run = True

    def SetArgs(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    def Start(self, millis=None, *args, **kwargs):
        if args:
            self._args = args
        if kwargs:
            self._kwargs = kwargs
        ThreadedTimer.Start(self, millis, True)

    Restart = Start


class BetterWindowDisabler(object):
    """Also disables child windows instead of only top level windows. This is
    actually needed under Mac OS X where disabling a top level window will
    not prevent interaction with its children.

    If toplevelparent is given, disable only this window and its child windows.

    """

    windows = set()

    def __init__(self, skip=None, toplevelparent=None, include_menus=False):
        self._windows = []
        self.skip = skip
        self.toplevelparent = toplevelparent
        self.include_menus = include_menus
        self.disable()

    def __del__(self):
        self.enable()

    def disable(self):
        self.enable(False)

    def enable(self, enable=True):
        if not enable:
            skip = self.skip or []
            if skip:
                if not isinstance(skip, (list, tuple)):
                    skip = [skip]
            if self.toplevelparent:
                toplevel = [self.toplevelparent]
            else:
                toplevel = list(wx.GetTopLevelWindows())
            for w in toplevel:
                if (
                    w
                    and w not in skip
                    and "Inspection" not in "%s" % w
                    and w not in BetterWindowDisabler.windows
                ):
                    self._windows.append(w)
                    # Selectively add children to our list of handled
                    # windows. This prevents a segfault with wxPython 4
                    # under macOS where GetAllChildren includes sub-controls
                    # of controls, like scrollbars etc.
                    for child in w.GetAllChildren(skip + toplevel):
                        if (
                            child
                            and isinstance(
                                child,
                                (
                                    wx.BitmapButton,
                                    wx.Button,
                                    wx.CheckBox,
                                    wx.Choice,
                                    wx.ComboBox,
                                    wx.ListBox,
                                    wx.ListCtrl,
                                    wx.RadioButton,
                                    wx.SpinCtrl,
                                    wx.Slider,
                                    wx.StaticText,
                                    wx.TextCtrl,
                                    wx.grid.Grid,
                                    floatspin.FloatSpin,
                                    GenButton,
                                    GradientButton,
                                    PlateButton,
                                ),
                            )
                            and child not in BetterWindowDisabler.windows
                        ):
                            # Don't disable panels, this can have weird side
                            # effects for contained controls
                            self._windows.append(child)
                    if self.include_menus:
                        menubar = w.GetMenuBar()
                        if menubar:
                            for menu, _label in menubar.GetMenus():
                                for item in menu.GetMenuItems():
                                    self._windows.append(item)

            def Enable(w_, enable=True):
                w_._BetterWindowDisabler_enabled = enable

            def Disable(w_):
                w_._BetterWindowDisabler_enabled = False

            for w in reversed(self._windows):
                BetterWindowDisabler.windows.add(w)
                enabled = w.IsEnabled()
                w.Enable(False)
                if hasattr(w, "Disable"):
                    w._BetterWindowDisabler_Disable = w.Disable
                    # w.Disable = types.MethodType(Disable, w, type(w))
                    w.Disable = types.MethodType(Disable, w)
                w._BetterWindowDisabler_Enable = w.Enable
                # w.Enable = types.MethodType(Enable, w, type(w))
                w.Enable = types.MethodType(Enable, w)
                w.Enable(enabled)
            return

        for w in self._windows:
            BetterWindowDisabler.windows.remove(w)
            if w:
                if hasattr(w, "_BetterWindowDisabler_Disable"):
                    w.Disable = w._BetterWindowDisabler_Disable
                if hasattr(w, "_BetterWindowDisabler_Enable"):
                    w.Enable = w._BetterWindowDisabler_Enable
                if hasattr(w, "_BetterWindowDisabler_enabled"):
                    w.Enable(w._BetterWindowDisabler_enabled)


class CustomGridCellEvent(CustomEvent):
    def __init__(self, typeId, object, row=-1, col=-1, window=None):
        CustomEvent.__init__(self, typeId, object, window)
        self.Row = row
        self.Col = col

    def GetRow(self):
        return self.Row

    def GetCol(self):
        return self.Col


class PopupMenu(object):
    """A collection of menus that has a wx.MenuBar-like interface."""

    def __init__(self, parent):
        self.Parent = parent
        self.TopLevelParent = parent.TopLevelParent
        self._menus = []
        self._enabledtop = {}

    def Append(self, menu, title):
        self._menus.append((menu, title))

    def EnableTop(self, pos, enable=True):
        self._enabledtop[pos] = enable

    def FindItemById(self, id):
        for menu, _label in self._menus:
            item = menu.FindItemById(id)
            if item:
                return item

    def FindMenu(self, title):
        for i, (_menu, label) in enumerate(self._menus):
            if title == label:
                return i
        return wx.NOT_FOUND

    def GetMenu(self, index):
        return self._menus[index][0]

    def GetMenuCount(self):
        return len(self._menus)

    def GetMenus(self):
        """Return menus.

        Returns:
            List[wx.Menu]: List of child menus.
        """
        return list(self._menus)

    def SetMenus(self, menus):
        """Set menus.

        Args:
            menus (List[wx.Menu]): A list of wx.Menu instances to set as a
                child of this menubar.
        """
        self._menus = []
        for menu, label in menus:
            self.Append((menu, label))

    Menus = property(GetMenus, SetMenus)

    def IsEnabledTop(self, pos):
        return self._enabledtop.get(pos, True)

    def SetMenuLabel(self, pos, label):
        self._menus[pos] = (self._menus[pos][0], label)

    def bind_keys(self):
        if sys.platform == "darwin":
            accels = self.get_accelerator_entries()
            self.TopLevelParent.SetAcceleratorTable(wx.AcceleratorTable(accels()))
        else:
            self.TopLevelParent.Bind(wx.EVT_CHAR_HOOK, self.key_handler)

    def get_accelerator_entries(self):
        accels = []
        for menu, _label in self._menus:
            for item in menu.MenuItems:
                accel = item.Accel
                if accel:
                    accel = wx.AcceleratorEntry(
                        accel.Flags, accel.KeyCode, accel.Command, item
                    )
                    accels.append(accel)
        return accels

    def key_handler(self, event):
        """Handle accelerator keys"""
        keycode = event.KeyCode
        flags = wx.ACCEL_NORMAL
        for key in ("ALT", "CMD", "CTRL", "SHIFT"):
            if wx.GetKeyState(getattr(wx, "WXK_" + key.replace("CTRL", "CONTROL"), -1)):
                flags |= getattr(wx, "ACCEL_" + key.upper())
        for menu, _label in self._menus:
            for item in menu.MenuItems:
                accel = item.Accel
                if accel and accel.KeyCode == keycode and accel.Flags == flags:
                    event = wx.CommandEvent(wx.wxEVT_COMMAND_MENU_SELECTED)
                    event.Id = item.Id
                    if item.Kind == wx.ITEM_RADIO:
                        event.SetInt(1)
                    elif item.Kind == wx.ITEM_CHECK:
                        event.SetInt(int(not item.Checked))
                    self.TopLevelParent.ProcessEvent(event)

    def popup(self):
        """Popup the list of menus (with actual menus as submenus)"""

        top_menu = wx.Menu()

        for menu, label in self._menus:
            top_menu.AppendSubMenu(menu, label)

        self.Parent.PopupMenu(top_menu)

        # Delete menuitems (not submenus)
        for item in top_menu.MenuItems:
            top_menu.Delete(item)

        # Now we can safely destroy the menu without affecting submenus
        top_menu.Destroy()


class FileDrop(wx.FileDropTarget):
    def __init__(self, drophandlers=None):
        wx.FileDropTarget.__init__(self)
        if drophandlers is None:
            drophandlers = {}
        self.drophandlers = drophandlers
        self.unsupported_handler = None

    def OnDropFiles(self, x, y, filenames):
        dialogs = get_dialogs()
        interactable = not hasattr(self, "parent") or (
            self.parent.Enabled and (not dialogs or self.parent in dialogs)
        )
        if not interactable:
            wx.Bell()
            return False
        self._files = []
        self._filenames = filenames

        for filename in filenames:
            name, ext = os.path.splitext(filename)
            if ext.lower() in self.drophandlers:
                self._files.append((ext.lower(), filename))

        if self._files:
            self._files.reverse()
            wx.CallLater(1, wx.CallAfter, self.process)
        elif self.unsupported_handler:
            wx.CallLater(1, wx.CallAfter, self.unsupported_handler)
        return False

    def process(self):
        ms = 1.0 / 60
        while self._files:
            if hasattr(self, "parent") and hasattr(self.parent, "worker"):
                while self.parent.worker.is_working():
                    wx.Yield()
                    sleep(ms)
                    if self.parent.worker.thread_abort:
                        return
            ext, filename = self._files.pop()
            self.drophandlers[ext](filename)


class IdFactory(object):
    """Inspired by wxPython 4 (Phoenix) wx.IdManager"""

    CurrentId = 100
    ReservedIds = set()

    @classmethod
    def NewId(cls):
        """Replacement for wx.NewId()"""
        start_id = cls.CurrentId

        while True:
            # Skip the part of IDs space that contains hard-coded values
            if cls.CurrentId == wx.ID_LOWEST:
                cls.CurrentId = wx.ID_HIGHEST + 1
            id = cls.CurrentId
            if id < 30095:
                cls.CurrentId += 1
            else:
                cls.CurrentId = 100
            if id not in cls.ReservedIds:
                break
            elif cls.CurrentId == start_id:
                raise RuntimeError(
                    "Error: Out of IDs. Recommend shutting down application."
                )

        cls.ReserveId(id)

        return id

    @classmethod
    def ReserveId(cls, id):
        cls.ReservedIds.add(id)

    @classmethod
    def UnreserveId(cls, id):
        cls.ReservedIds.remove(id)



# -*- coding: utf-8 -*-c
"""demjson 1.3 compatibility module"""

from io import StringIO
import json
import sys


DEBUG = False


def decode(txt, strict=False, encoding=None, **kw):
    """Decodes a JSON-encoded string into a Python object.

    If 'strict' is set to True, then only strictly-conforming JSON
    output will be produced.  Note that this means that some types
    of values may not be convertable and will result in a
    JSONEncodeError exception.

    The input string can be either a python string or a python unicode
    string.  If it is already a unicode string, then it is assumed
    that no character set decoding is required.

    However, if you pass in a non-Unicode text string (i.e., a python
    type 'str') then an attempt will be made to auto-detect and decode
    the character encoding.  This will be successful if the input was
    encoded in any of UTF-8, UTF-16 (BE or LE), or UTF-32 (BE or LE),
    and of course plain ASCII works too.

    Note though that if you know the character encoding, then you
    should convert to a unicode string yourself, or pass it the name
    of the 'encoding' to avoid the guessing made by the auto
    detection, as with

        python_object = demjson.decode( input_bytes, encoding='utf8' )

    Optional keywords arguments are ignored.

    """

    if not strict:
        # Remove comments
        io = StringIO()
        escape = False
        prev = None
        expect_comment = False
        in_comment = False
        comment_multiline = False
        in_quote = False
        write = True
        for c in txt:
            if DEBUG:
                sys.stdout.write(c)
            write = True
            if c == "\\":
                if DEBUG:
                    sys.stdout.write("<ESCAPE>")
                escape = True
            elif escape:
                if DEBUG:
                    sys.stdout.write("</ESCAPE>")
                escape = False
            else:
                if not in_quote:
                    if c == "/":
                        if expect_comment:
                            if DEBUG:
                                sys.stdout.write("<COMMENT>")
                            in_comment = True
                            comment_multiline = False
                            expect_comment = False
                        elif in_comment and prev == "*":
                            if DEBUG:
                                sys.stdout.write("</MULTILINECOMMENT>")
                            in_comment = False
                            comment_multiline = False
                            write = False
                        elif not in_comment:
                            if DEBUG:
                                sys.stdout.write("<EXPECT_COMMENT>")
                            expect_comment = True
                    elif c == "*":
                        if expect_comment:
                            if DEBUG:
                                sys.stdout.write("<MULTILINECOMMENT>")
                            in_comment = True
                            comment_multiline = True
                            expect_comment = False
                    elif expect_comment:
                        if DEBUG:
                            sys.stdout.write("</EXPECT_COMMENT>")
                        expect_comment = False
                if c == "\n":
                    if in_comment and not comment_multiline:
                        if DEBUG:
                            sys.stdout.write("</COMMENT>")
                        in_comment = False
                        write = False
                elif c == '"' and not in_comment:
                    if in_quote:
                        if DEBUG:
                            sys.stdout.write("</QUOTE>")
                        in_quote = False
                    else:
                        if DEBUG:
                            sys.stdout.write("<QUOTE>")
                        in_quote = True
            if write and not expect_comment and not in_comment:
                io.write(c)
            prev = c
        txt = io.getvalue()
        if DEBUG:
            sys.stdout.write("\n")
            print("JSON:", txt)

    return json.loads(txt, encoding=encoding, strict=strict)


def encode(obj, strict=False, compactly=True, escape_unicode=False, encoding=None):
    """Encodes a Python object into a JSON-encoded string.

    'strict' is ignored.

    If 'compactly' is set to True, then the resulting string will
    have all extraneous white space removed; if False then the
    string will be "pretty printed" with whitespace and indentation
    added to make it more readable.

    If 'escape_unicode' is set to True, then all non-ASCII characters
    will be represented as a unicode escape sequence; if False then
    the actual real unicode character will be inserted.

    If no encoding is specified (encoding=None) then the output will
    either be a Python string (if entirely ASCII) or a Python unicode
    string type.

    However if an encoding name is given then the returned value will
    be a python string which is the byte sequence encoding the JSON
    value.  As the default/recommended encoding for JSON is UTF-8,
    you should almost always pass in encoding='utf8'.

    """

    if compactly:
        indent = None
        separators = (",", ":")
    else:
        indent = 2
        separators = (",", ": ")

    ensure_ascii = escape_unicode or encoding is not None

    return json.dumps(
        obj,
        ensure_ascii=ensure_ascii,
        indent=indent,
        separators=separators,
        encoding=encoding or "utf-8",
    )



# -*- coding: utf-8 -*-


import math
import os
import struct
import time
import zlib

from DisplayCAL.meta import name as appname, version
from DisplayCAL.util_str import safe_str


TIFF_TAG_TYPE_BYTE = 1
TIFF_TAG_TYPE_ASCII = 2
TIFF_TAG_TYPE_WORD = 3
TIFF_TAG_TYPE_DWORD = 4
TIFF_TAG_TYPE_RATIONAL = 5  # 2 DWORDs


def tiff_get_header(w, h, samples_per_pixel, bitdepth):
    # Very helpful: http://www.fileformat.info/format/tiff/corion.htm

    header = []
    header.append(b"MM\0*")  # Note: We use big-endian byte order

    # Offset of image directory
    header.append(b"\0\0\0\x08")

    pixelcount = w * h * samples_per_pixel
    if bitdepth == 16:
        bytecount = pixelcount * 2
    else:
        bytecount = pixelcount

    # Image file directory (IFD)

    # PhotometricInterpretation
    if samples_per_pixel == 3:
        pmi = 2  # RGB
    else:
        pmi = 5  # Separated (usually CMYK)

    # Tag, type, length, offset or data, is data (otherwise offset)
    if w > 65535:
        tag_type_w = TIFF_TAG_TYPE_DWORD
    else:
        tag_type_w = TIFF_TAG_TYPE_WORD
    if h > 65535:
        tag_type_h = TIFF_TAG_TYPE_DWORD
    else:
        tag_type_h = TIFF_TAG_TYPE_WORD
    ifd = [
        (0x100, tag_type_w, 1, w, True),  # ImageWidth
        (0x101, tag_type_h, 1, h, True),  # ImageLength
        (0x106, TIFF_TAG_TYPE_WORD, 1, pmi, True),  # PhotometricInterpretation
        (0x115, TIFF_TAG_TYPE_WORD, 1, samples_per_pixel, True),  # SamplesPerPixel
        (0x117, TIFF_TAG_TYPE_DWORD, 1, bytecount, True),  # StripByteCounts
    ]
    # BitsPerSample
    ifd.append((0x102, TIFF_TAG_TYPE_WORD, 3, 10 + (len(ifd) + 2) * 12 + 4, False))
    # StripOffsets
    ifd.append((0x111, TIFF_TAG_TYPE_WORD, 1, 10 + (len(ifd) + 1) * 12 + 4 + 6, True))

    ifd.sort()  # Must be ascending order!

    header.append(struct.pack(">H", len(ifd)))  # Number of entries

    for tag, tagtype, length, payload, is_data in ifd:
        header.append(struct.pack(">H", tag))
        header.append(struct.pack(">H", tagtype))
        header.append(struct.pack(">I", length))
        if is_data and tagtype == 3:
            # A word left-aligned in a dword
            header.append(struct.pack(">H", payload))
            header.append(b"\0\0")
        else:
            header.append(struct.pack(">I", payload))

    # PlanarConfiguration default is 1 = RGBRGBRGB...

    # End of IFD
    header.append(b"\0" * 4)

    # BitsPerSample (6 bytes)
    header.append(struct.pack(">H", bitdepth) * 3)

    return b"".join(header)


def write(
    data, stream_or_filename, bitdepth=16, format=None, dimensions=None, extrainfo=None
):
    Image(data, bitdepth, extrainfo).write(stream_or_filename, format, dimensions)


def write_rgb_clut(stream_or_filename, clutres=33, bitdepth=16, format=None):
    clut = []
    for R in range(clutres):
        for G in range(clutres):
            clut.append([])
            for B in range(clutres):
                RGB = [v * (1.0 / (clutres - 1)) for v in (R, G, B)]
                clut[-1].append([v * (2**bitdepth - 1) for v in RGB])
    write(clut, stream_or_filename, bitdepth, format)


class Image(object):
    """Write 8 or 16 bit image files in DPX, PNG or TIFF format.

    Writing of single color images is highly optimized when using a single
    pixel as image data and setting dimensions explicitly.

    """

    def __init__(self, data, bitdepth=16, extrainfo=None):
        self.bitdepth = bitdepth
        self.data = data
        self.extrainfo = extrainfo or {}

    def _pack(self, n):
        n = int(round(n))
        if self.bitdepth == 16:
            data = struct.pack(">H", n)
        elif self.bitdepth == 8:
            data = n.to_bytes(1, "big")
        else:
            raise ValueError("Unsupported bitdepth: %r" % self.bitdepth)
        return data

    def _write_dpx(self, stream, dimensions=None):
        # Very helpful: http://www.fileformat.info/format/dpx/egff.htm
        # http://www.simplesystems.org/users/bfriesen/dpx/S268M_Revised.pdf

        # Generic file header (768 bytes)
        stream.write(b"SDPX")  # Magic number
        stream.write(struct.pack(">I", 8192))  # Offset to image data
        stream.write(b"V2.0\0\0\0\0")  # ASCII version

        # Optimize for single color
        optimize = len(self.data) == 1 and len(self.data[0]) == 1 and dimensions

        # Image data
        imgdata = []
        # 10-bit code adapted from GraphicsMagick dpx.c:WriteSamples
        if self.bitdepth == 10:
            shifts = (22, 12, 2)  # RGB
        for _i, scanline in enumerate(self.data):
            if self.bitdepth == 10:
                packed = []
                for RGB in scanline:
                    packed_u32 = 0
                    for datum, sample in enumerate(RGB):
                        packed_u32 |= sample << shifts[datum]
                    packed.append(struct.pack(">I", packed_u32))
                scanline = b"".join(packed)
            else:
                scanline = b"".join(
                    b"".join(self._pack(v) for v in RGB) for RGB in scanline
                )
            if not optimize:
                # Pad lines with binary zeros so they end on 4-byte boundaries
                scanline = scanline.ljust(
                    int(math.ceil(len(scanline) / 4.0)) * 4, b"\0"
                )
            imgdata.append(scanline)
        imgdata = "".join(imgdata)
        if optimize:
            # Optimize for single color
            imgdata *= dimensions[0]
            # Pad lines with binary zeros so they end on 4-byte boundaries
            imgdata = imgdata.ljust(int(math.ceil(len(imgdata) / 4.0)) * 4, b"\0")
            imgdata *= dimensions[1]
            w, h = dimensions
        else:
            w, h = len(self.data[0]), len(self.data)

        # Generic file header (cont.)
        stream.write(struct.pack(">I", 8192 + len(imgdata)))  # File size
        stream.write(b"\0\0\0\1")  # DittoKey (1 = not same as previous frame)
        stream.write(
            struct.pack(">I", 768 + 640 + 256)
        )  # Generic section header length
        stream.write(
            struct.pack(">I", 256 + 128)
        )  # Industry-specific section header length
        stream.write(struct.pack(">I", 0))  # User-defined data length
        stream.write(safe_str(stream.name or b"").ljust(100, b"\0")[-100:])  # File name
        # Date & timestamp
        tzoffset = round(
            (time.mktime(time.localtime()) - time.mktime(time.gmtime())) / 60.0 / 60.0
        )
        if tzoffset < 0:
            tzoffset = b"%.2i" % tzoffset
        else:
            tzoffset = b"+%.2i" % tzoffset
        stream.write(
            time.strftime("%Y:%m:%d:%H:%M:%S").encode() + tzoffset.encode() + b"\0\0"
        )
        stream.write(
            safe_str(b"%s %s" % (appname, version)).ljust(100, b"\0")
        )  # Creator
        stream.write(b"\0" * 200)  # Project
        stream.write(b"\0" * 200)  # Copyright
        stream.write(b"\xff" * 4)  # EncryptKey 0xffffffff = not encrypted
        stream.write(b"\0" * 104)  # Reserved

        # Generic image header (640 bytes)
        stream.write(b"\0\0")  # Orientation 0 = left to right, top to bottom
        stream.write(b"\0\1")  # Number of image elements
        stream.write(struct.pack(">I", w))  # Pixels per line
        stream.write(struct.pack(">I", h))  # Lines per image element

        # Generic image header - image element
        stream.write(b"\0" * 4)  # 0 = unsigned data
        stream.write(b"\0" * 4)  # Reference low data code value
        stream.write(b"\xff" * 4)  # Reference low quantity
        stream.write(
            struct.pack(">I", 2**self.bitdepth - 1)
        )  # Reference high data code value
        stream.write(b"\xff" * 4)  # Reference high quantity
        stream.write(chr(50).encode())  # Descriptor 50 = RGB
        stream.write(b"\2")  # Transfer 2 = linear
        stream.write(b"\2")  # Colorimetric 2 = not applicable
        stream.write(chr(self.bitdepth).encode())  # BitSize
        stream.write(b"\0\1")  # Packing 1 = filled 32-bit words
        stream.write(b"\0\0")  # Encoding 0 = not encoded
        stream.write(struct.pack(">I", 8192))  # Image data offset
        stream.write(b"\0" * 4)  # End of line padding
        stream.write(b"\0" * 4)  # End of image padding
        stream.write(b"RGB / Linear".ljust(32, b"\0"))  # Description

        # Seven additional unused image elements
        stream.write(b"\0" * 72 * 7)

        # Generic image header (cont.)
        stream.write(b"\0" * 52)  # Reserved

        # Generic image source header (256 bytes)
        sw, sh = [
            self.extrainfo.get("original_" + dim, locals()[dim[0]])
            for dim in ("width", "height")
        ]
        # X offset
        stream.write(struct.pack(">I", self.extrainfo.get("offset_x", (sw - w) / 2)))
        # Y offset
        stream.write(struct.pack(">I", self.extrainfo.get("offset_y", (sh - h) / 2)))
        # X center
        stream.write(struct.pack(">f", self.extrainfo.get("center_x", sw / 2.0)))
        # Y center
        stream.write(struct.pack(">f", self.extrainfo.get("center_y", sh / 2.0)))
        stream.write(struct.pack(">I", sw))  # X original size
        stream.write(struct.pack(">I", sh))  # Y original size
        stream.write(b"\0" * 100)  # Source image file name
        stream.write(b"\0" * 24)  # Source image date & timestamp
        stream.write(b"\0" * 32)  # Input device name
        stream.write(b"\0" * 32)  # Input device serial number
        stream.write(b"\0" * 2 * 4)  # Border
        stream.write(b"\0\0\0\1" * 2)  # Pixel aspect ratio
        stream.write(b"\xff" * 4)  # X scanned size
        stream.write(b"\xff" * 4)  # Y scanned size
        stream.write(b"\0" * 20)  # Reserved

        # Industry-specific film info header (256 bytes)
        stream.write(b"\0" * 2)  # Film mfg. ID code
        stream.write(b"\0" * 2)  # Film type
        stream.write(b"\0" * 2)  # Offset in perfs
        stream.write(b"\0" * 6)  # Prefix
        stream.write(b"\0" * 4)  # Count
        stream.write(b"\0" * 32)  # Format
        # Frame position in sequence
        stream.write(struct.pack(">I", self.extrainfo.get("frame_position", 2**32 - 1)))
        # Sequence length
        stream.write(
            struct.pack(">I", self.extrainfo.get("sequence_length", 2**32 - 1))
        )
        # Held count
        stream.write(struct.pack(">I", self.extrainfo.get("held_count", 1)))
        # Frame rate of original
        if "frame_rate" in self.extrainfo:
            stream.write(struct.pack(">f", self.extrainfo["frame_rate"]))
        else:
            stream.write(b"\xff" * 4)
        # Shutter angle of camera in degrees
        stream.write(b"\xff" * 4)
        stream.write(b"\0" * 32)  # Frame identification - e.g. keyframe
        stream.write(b"\0" * 100)  # Slate
        stream.write(b"\0" * 56)  # Reserved

        # Industry-specific TV info header (128 bytes)
        # SMPTE time code
        stream.write(
            b"".join(
                chr(int(str(v), 16)).encode()
                for v in self.extrainfo.get("timecode", [b"ff"] * 4)
            )
        )
        stream.write(b"\xff" * 4)  # User bits
        stream.write(b"\xff")  # Interlace
        stream.write(b"\xff")  # Field number
        stream.write(b"\xff")  # Video signal standard
        stream.write(b"\0")  # Zero for byte alignment
        stream.write(b"\xff" * 4)  # H sampling rate Hz
        stream.write(b"\xff" * 4)  # V sampling rate Hz
        # Temporal sampling or frame rate Hz
        if "frame_rate" in self.extrainfo:
            stream.write(struct.pack(">f", self.extrainfo["frame_rate"]))
        else:
            stream.write(b"\xff" * 4)
        stream.write(b"\xff" * 4)  # Time offset in ms from sync to 1st pixel
        stream.write(b"\xff" * 4)  # Gamma
        stream.write(b"\xff" * 4)  # Black level code value
        stream.write(b"\xff" * 4)  # Black gain
        stream.write(b"\xff" * 4)  # Breakpoint
        stream.write(b"\xff" * 4)  # Reference white level code value
        stream.write(b"\xff" * 4)  # Integration time in s
        stream.write(b"\0" * 76)  # Reserved

        # Padding so image data begins at 8K boundary
        stream.write("\0" * 6144)

        # Write image data
        stream.write(imgdata)

    def _write_png(self, stream, dimensions=None):
        # Header
        stream.write(b"\x89PNG\r\n\x1a\n")
        # IHDR image header length
        stream.write(struct.pack(">I", 13))
        # IHDR image header chunk type
        ihdr = [b"IHDR"]
        # Optimize for single color
        optimize = len(self.data) == 1 and len(self.data[0]) == 1 and dimensions
        # IHDR: width, height
        if optimize:
            w, h = dimensions
        else:
            w, h = len(self.data[0]), len(self.data)
        ihdr.extend([struct.pack(">I", w), struct.pack(">I", h)])
        # IHDR: Bit depth
        ihdr.append(self.bitdepth.to_bytes(1, "big"))
        # IHDR: Color type 2 (truecolor)
        ihdr.append(b"\2")
        # IHDR: Compression method 0 (deflate)
        ihdr.append(b"\0")
        # IHDR: Filter method 0 (adaptive)
        ihdr.append(b"\0")
        # IHDR: Interlace method 0 (none)
        ihdr.append(b"\0")
        ihdr = b"".join(ihdr)
        stream.write(ihdr)
        stream.write(struct.pack(">I", zlib.crc32(ihdr) & 0xFFFFFFFF))
        # IDAT image data chunk type
        imgdata = []
        for _i, scanline in enumerate(self.data):
            # Add a scanline, filter type 0
            imgdata.append(b"\0")
            for RGB in scanline:
                RGB = b"".join(self._pack(v) for v in RGB)
                if optimize:
                    RGB *= dimensions[0]
                imgdata.append(RGB)
        imgdata = b"".join(imgdata)
        if optimize:
            imgdata *= dimensions[1]
        imgdata = zlib.compress(imgdata, 9)
        stream.write(struct.pack(">I", len(imgdata)))
        idat = [b"IDAT"]
        idat.append(imgdata)
        idat = b"".join(idat)
        stream.write(idat)
        stream.write(struct.pack(">I", zlib.crc32(idat) & 0xFFFFFFFF))
        # IEND chunk
        stream.write(b"\0" * 4)
        stream.write(b"IEND")
        stream.write(struct.pack(">I", zlib.crc32(b"IEND") & 0xFFFFFFFF))

    def _write_tiff(self, stream, dimensions=None):
        # Very helpful: http://www.fileformat.info/format/tiff/corion.htm

        # Image data
        if len(self.data) == 1 and len(self.data[0]) == 1 and dimensions:
            # Optimize for single color
            w, h = dimensions
            imgdata = list(self.data)
            imgdata[0] *= w
            imgdata *= h
        else:
            imgdata = self.data
            w, h = len(self.data[0]), len(self.data)

        samples_per_pixel = len(self.data[0][0])

        # Header
        stream.write(tiff_get_header(w, h, samples_per_pixel, self.bitdepth))

        # Write image data
        for _i, scanline in enumerate(imgdata):
            for sample in scanline:
                stream.write(b"".join(self._pack(v) for v in sample))

    def write(self, stream_or_filename, format=None, dimensions=None):
        if not format:
            if isinstance(stream_or_filename, str):
                format = os.path.splitext(stream_or_filename)[1].lstrip(".").upper()
                if format == "TIF":
                    format += "F"
            else:
                format = "PNG"
        if not hasattr(self, "_write_" + format.lower()):
            raise ValueError("Unsupported format: %r" % format)
        if isinstance(stream_or_filename, str):
            stream = open(stream_or_filename, "wb")
        else:
            stream = stream_or_filename
        with stream:
            getattr(self, "_write_" + format.lower())(stream, dimensions)



# -*- coding: utf-8 -*-

import os
import sys

from DisplayCAL.meta import name as appname
from DisplayCAL.util_os import launch_file, make_win32_compatible_long_path, waccess
from DisplayCAL import config
from DisplayCAL import localization as lang
from DisplayCAL import x3dom

gui = "wx" in sys.modules

if gui:
    from DisplayCAL.worker import Worker, show_result_dialog
    from DisplayCAL.wxaddons import wx
    from DisplayCAL.wxfixes import GenBitmapButton as BitmapButton
    from DisplayCAL.wxwindows import BaseApp, BaseFrame, FileDrop

    class VRML2X3DFrame(BaseFrame):
        def __init__(self, html, embed, view, force, cache):
            BaseFrame.__init__(
                self,
                None,
                wx.ID_ANY,
                lang.getstr("vrml_to_x3d_converter"),
                style=wx.DEFAULT_FRAME_STYLE & ~(wx.MAXIMIZE_BOX | wx.RESIZE_BORDER),
                name="vrml2x3dframe",
            )
            self.SetIcons(
                config.get_icon_bundle(
                    [256, 48, 32, 16], "%s-VRML-to-X3D-converter" % appname
                )
            )
            self.Bind(wx.EVT_CLOSE, self.OnClose)
            self.cache = cache
            self.embed = embed
            self.force = force
            self.html = html
            self.worker = Worker(self)
            sizer = wx.BoxSizer(wx.HORIZONTAL)
            self.SetSizer(sizer)
            panel = wx.Panel(self)
            sizer.Add(panel)
            panelsizer = wx.BoxSizer(wx.HORIZONTAL)
            panel.SetSizer(panelsizer)
            self.btn = BitmapButton(
                panel,
                wx.ID_ANY,
                config.geticon(256, "3d-primitives"),
                style=wx.NO_BORDER,
            )
            self.btn.SetToolTipString(lang.getstr("file.select"))
            self.btn.Bind(
                wx.EVT_BUTTON,
                lambda event: vrmlfile2x3dfile(
                    None,
                    html=html,
                    embed=embed,
                    view=view,
                    force=force,
                    cache=cache,
                    worker=self.worker,
                ),
            )
            self.droptarget = FileDrop(self)
            vrml_drop_handler = lambda vrmlpath: vrmlfile2x3dfile(
                vrmlpath,
                html=html,
                embed=embed,
                view=view,
                force=force,
                cache=cache,
                worker=self.worker,
            )
            self.droptarget.drophandlers = {
                ".vrml": vrml_drop_handler,
                ".vrml.gz": vrml_drop_handler,
                ".wrl": vrml_drop_handler,
                ".wrl.gz": vrml_drop_handler,
                ".wrz": vrml_drop_handler,
            }
            self.btn.SetDropTarget(self.droptarget)
            panelsizer.Add(self.btn, flag=wx.ALL, border=12)
            self.Fit()
            self.SetMinSize(self.GetSize())
            self.SetMaxSize(self.GetSize())

        def OnClose(self, event):
            # Hide first (looks nicer)
            self.Hide()
            # Need to use CallAfter to prevent hang under Windows if minimized
            wx.CallAfter(self.Destroy)

        def get_commands(self):
            return self.get_common_commands() + [
                "VRML-to-X3D-converter [filename...]",
                "load <filename...>",
            ]

        def process_data(self, data):
            if data[0] in ("VRML-to-X3D-converter", "load"):
                if self.IsIconized():
                    self.Restore()
                self.Raise()
                if len(data) > 1:
                    self.droptarget.OnDropFiles(0, 0, data[1:])
                return "ok"
            return "invalid"


def main():
    if "--help" in sys.argv[1:] or (not sys.argv[1:] and not gui):
        print("Convert VRML file to X3D")
        print("Author: Florian Hoech, licensed under the GPL version 3")
        print("Usage: %s [OPTION]... FILE..." % os.path.basename(sys.argv[0]))
        print("The output is written to FILENAME.x3d(.html)")
        print("")
        print(
            "  --embed      Embed viewer components in HTML instead of referencing them"
        )
        print("  --force      Force fresh download of viewer components")
        print(
            "  --no-cache   Don't use viewer components cache (only uses existing cache if"
        )
        print("               embedding components, can be overridden with --force)")
        if gui:
            print("  --no-gui     Don't use GUI (console mode)")
        print("  --no-html    Don't generate HTML file")
        print("  --view       View the generated file (if no GUI)")
        if not gui:
            print("  --batch      Don't pause after processing")
        print("  FILE         Filename of VRML file to convert")
        if gui:
            return
    if gui:
        config.initcfg("VRML-to-X3D-converter")
        lang.init()
        lang.update_defaults()
    cache = "--no-cache" not in sys.argv[1:]
    embed = "--embed" in sys.argv
    force = "--force" in sys.argv
    html = "--no-html" not in sys.argv[1:]
    if not gui:
        result = None
        view = "--view" in sys.argv[1:]
        for arg in sys.argv[1:]:
            if not arg.startswith("--"):
                result = vrmlfile2x3dfile(
                    str(arg),
                    html=html,
                    embed=embed,
                    view=view,
                    force=force,
                    cache=cache,
                    gui=gui,
                )
        if result is None:
            print("No filename given.")
        if (
            sys.stdout
            and hasattr(sys.stdout, "isatty")
            and sys.stdout.isatty()
            and "--batch" not in sys.argv[1:]
        ):
            input("Press RETURN to exit")
        sys.exit(int(not result))
    else:
        view = "--no-view" not in sys.argv[1:]
        app = BaseApp(0)
        app.TopWindow = VRML2X3DFrame(html, embed, view, force, cache)
        if sys.platform == "darwin":
            app.TopWindow.init_menubar()
        wx.CallLater(1, _main, app)
        app.MainLoop()


def _main(app):
    app.TopWindow.listen()
    app.process_argv()
    app.TopWindow.Show()


def vrmlfile2x3dfile(
    vrmlpath=None,
    x3dpath=None,
    html=True,
    embed=False,
    view=False,
    force=False,
    cache=True,
    worker=None,
    gui=True,
):
    """Convert VRML to HTML. Output is written to <vrmlfilename>.x3d.html
    unless you set x3dpath to desired output path, or False to be prompted
    for an output path."""
    while not vrmlpath or not os.path.isfile(vrmlpath):
        if not gui:
            if not vrmlpath or vrmlpath.startswith("--"):
                print("No filename given.")
            else:
                print("%r is not a file." % vrmlpath)
            return False
        if not wx.GetApp():
            app = BaseApp(0)
        defaultDir, defaultFile = config.get_verified_path("last_vrml_path")
        dlg = wx.FileDialog(
            None,
            lang.getstr("file.select"),
            defaultDir=defaultDir,
            defaultFile=defaultFile,
            wildcard=lang.getstr("filetype.vrml")
            + "|*.vrml;*.vrml.gz;*.wrl.gz;*.wrl;*.wrz",
            style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST,
        )
        dlg.Center(wx.BOTH)
        result = dlg.ShowModal()
        vrmlpath = dlg.GetPath()
        dlg.Destroy()
        if result != wx.ID_OK:
            return
        config.setcfg("last_vrml_path", vrmlpath)
        config.writecfg(module="VRML-to-X3D-converter", options=("last_vrml_path",))
    filename, ext = os.path.splitext(vrmlpath)
    if x3dpath is None:
        x3dpath = filename + ".x3d"
    if x3dpath:
        dirname = os.path.dirname(x3dpath)
    while not x3dpath or not waccess(dirname, os.W_OK):
        if not gui:
            if not x3dpath:
                print("No HTML output filename given.")
            else:
                print("%r is not writable." % dirname)
            return False
        if not wx.GetApp():
            app = BaseApp(0)
        if x3dpath:
            defaultDir, defaultFile = os.path.split(x3dpath)
        else:
            defaultFile = os.path.basename(filename) + ".x3d"
        dlg = wx.FileDialog(
            None,
            lang.getstr("error.access_denied.write", dirname),
            defaultDir=defaultDir,
            defaultFile=defaultFile,
            wildcard=lang.getstr("filetype.x3d") + "|*.x3d",
            style=wx.SAVE | wx.FD_OVERWRITE_PROMPT,
        )
        dlg.Center(wx.BOTH)
        result = dlg.ShowModal()
        dlg.Destroy()
        if result != wx.ID_OK:
            return
        x3dpath = dlg.GetPath()
        dirname = os.path.dirname(x3dpath)
    vrmlpath, x3dpath = [str(path) for path in (vrmlpath, x3dpath)]
    if sys.platform == "win32":
        vrmlpath = make_win32_compatible_long_path(vrmlpath)
        x3dpath = make_win32_compatible_long_path(x3dpath)
    if html:
        finalpath = x3dpath + ".html"
        if sys.platform == "win32":
            finalpath = make_win32_compatible_long_path(finalpath)
            x3dpath = finalpath[:-5]
    else:
        finalpath = x3dpath
    if worker:
        worker.clear_cmd_output()
        worker.start(
            lambda result: (
                show_result_dialog(result, wx.GetApp().GetTopWindow())
                if isinstance(result, Exception)
                else result and view and launch_file(finalpath)
            ),
            x3dom.vrmlfile2x3dfile,
            wargs=(vrmlpath, x3dpath, html, embed, force, cache, worker),
            progress_title=lang.getstr("vrml_to_x3d_converter"),
            progress_start=1,
            resume=worker.progress_wnd and worker.progress_wnd.IsShownOnScreen(),
            fancy=False,
        )
    else:
        result = x3dom.vrmlfile2x3dfile(
            vrmlpath, x3dpath, html, embed, force, cache, None
        )
        if not isinstance(result, Exception) and result:
            if view:
                launch_file(finalpath)
        else:
            return False
    return True


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
import os
import pathlib
import shutil
import subprocess
import sys
import tarfile

import pytest
import tempfile

import DisplayCAL
from DisplayCAL.config import setcfg
from DisplayCAL.worker import get_argyll_latest_version


@pytest.fixture(scope="module")
def data_files():
    """generates data file list"""
    #  test/data
    extensions = ["*.txt", "*.tsv", "*.lin", "*.cal", "*.ti1", "*.ti3", "*.icc"]

    displaycal_parent_dir = pathlib.Path(DisplayCAL.__file__).parent
    search_paths = [
        displaycal_parent_dir,
        displaycal_parent_dir / "presets",
        displaycal_parent_dir / "ti1",
        displaycal_parent_dir.parent / "misc" / "ti3",
        displaycal_parent_dir.parent / "tests" / "data",
        displaycal_parent_dir.parent / "tests" / "data" / "sample",
        displaycal_parent_dir.parent / "tests" / "data" / "sample" / "issue129",
        displaycal_parent_dir.parent / "tests" / "data" / "sample" / "issue268",
        displaycal_parent_dir.parent / "tests" / "data" / "icc",
    ]
    d_files = {}
    for path in search_paths:
        for extension in extensions:
            # add files from DisplayCal/presets folder
            for element in path.glob(extension):
                d_files[element.name] = element

    yield d_files


@pytest.fixture(scope="module")
def data_path():
    """Return the tests/data folder path."""
    displaycal_parent_dir = pathlib.Path(DisplayCAL.__file__).parent
    return displaycal_parent_dir.parent / "tests" / "data"


@pytest.fixture(scope="module")
def argyll():
    """Setup ArgyllCMS.

    This will search for ArgyllCMS binaries under ``.local/bin/Argyll*/bin`` and if it
    can not find it, it will download from the source.
    """
    argyll_version = get_argyll_latest_version()
    argyll_download_url = {
        "win32": f"https://www.argyllcms.com/Argyll_V{argyll_version}_win64_exe.zip",
        "darwin": f"https://www.argyllcms.com/Argyll_V{argyll_version}_osx10.6_x86_64_bin.tgz",
        "linux": f"https://www.argyllcms.com/Argyll_V{argyll_version}_linux_x86_64_bin.tgz",
    }

    # first look in to ~/local/bin/ArgyllCMS
    home = pathlib.Path().home()
    argyll_search_paths = [
        home / ".local" / "bin" / "Argyll" / "bin",
        home / ".local" / "bin" / f"Argyll_V{argyll_version}" / "bin",
    ]

    argyll_path = None
    for path in argyll_search_paths:
        if path.is_dir():
            argyll_path = path
            setcfg("argyll.dir", str(argyll_path.absolute()))
            break

    if argyll_path:
        yield argyll_path
        return

    # apparently argyll has not been found
    # download from source
    url = argyll_download_url[sys.platform]

    argyll_temp_path = tempfile.mkdtemp()
    # store current working directory
    current_working_directory = os.getcwd()

    # change dir to argyll temp path
    os.chdir(argyll_temp_path)

    tar_file_name = "Argyll.tgz"
    if not os.path.exists(tar_file_name):
        print(f"Downloading: {tar_file_name}")
        # Download the tar file if it doesn't already exist
        subprocess.call(["/usr/bin/curl", url, "-o", tar_file_name])
    else:
        print(f"Tar file already exists: {tar_file_name}")
        print("Not downloading it again!")

    print(f"Decompressing Tarfile: {tar_file_name}")
    with tarfile.open(tar_file_name) as tar:
        tar.extractall()

    def cleanup():
        # cleanup the test
        shutil.rmtree(argyll_temp_path)
        os.chdir(current_working_directory)

    argyll_path = pathlib.Path(argyll_temp_path) / f"Argyll_V{argyll_version}" / "bin"
    print(f"argyll_path: {argyll_path}")
    if argyll_path.is_dir():
        setcfg("argyll.dir", str(argyll_path.absolute()))
        yield argyll_path
        cleanup()
    else:
        cleanup()
        pytest.skip("ArgyllCMS can not be setup!")


@pytest.fixture(scope="function")
def random_icc_profile():
    """Create a random ICCProfile suitable for modification."""
    import tempfile
    from DisplayCAL import colormath
    from DisplayCAL import ICCProfile

    rec709_gamma18 = list(colormath.get_rgb_space("Rec. 709"))
    icc_profile = ICCProfile.ICCProfile.from_rgb_space(
        rec709_gamma18, b"Rec. 709 gamma 1.8"
    )
    icc_profile_path = tempfile.mktemp(suffix=".icc")
    icc_profile.write(icc_profile_path)

    yield icc_profile, icc_profile_path

    # clean the file
    os.remove(icc_profile_path)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sys
from time import strftime


if __name__ == "__main__":
    if len(sys.argv) in (3, 4):
        curve = []
        start = float(sys.argv[1]) / 100
        end = float(sys.argv[2]) / 100
        steps = 256
        if len(sys.argv) > 3:
            steps = int(sys.argv[3])
        if start == end:
            print("ERROR: start == end")
        else:
            step = (end - start) / (steps - 1)
            i = start
            for _x in range(steps):
                curve += [i]
                i += step
            if curve[-1] != end:
                curve += [end]
            print("CAL    ")
            print("")
            print('DESCRIPTOR "Argyll Device Calibration Curves"')
            print('ORIGINATOR "lincal.py"')
            print('CREATED "%s"' % strftime("%a %b %d %H:%I:%S %Y"))
            print('KEYWORD "DEVICE_CLASS"')
            print('DEVICE_CLASS "DISPLAY"')
            print('KEYWORD "COLOR_REP"')
            print('COLOR_REP "RGB"')
            print("")
            print('KEYWORD "RGB_I"')
            print("NUMBER_OF_FIELDS 4")
            print("BEGIN_DATA_FORMAT")
            print("RGB_I RGB_R RGB_G RGB_B")
            print("END_DATA_FORMAT")
            print("")
            print("NUMBER_OF_SETS %i" % steps)
            print("BEGIN_DATA")
            for i in range(steps):
                print(
                    " ".join(
                        [
                            str(round(col, 7 if col < 0.1 else 6)).ljust(7, "0")
                            for col in [(1.0 / (steps - 1)) * i] + [curve[i]] * 3
                        ]
                    )
                )
            print("END_DATA")
    else:
        print("Usage: %s start end steps" % os.path.basename(__file__))
        print(" start, end = integer between 0 and 100")
        print(" steps = number of steps (default 256)")



# knownpaths.py (https://gist.github.com/mkropat/7550097)
#
# The MIT License (MIT)
#
# Copyright (c) 2014 Michael Kropat
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# I explicitly permit the knownpaths.py source code to be licensed by anyone
# under the terms of the GNU GPL v3, as an alternative to the X11/MIT license.

# Python 2.6/2.7 compatibility
import ctypes
import sys
from ctypes import windll, wintypes
from uuid import UUID


class GUID(ctypes.Structure):  # [1]
    _fields_ = [
        ("Data1", wintypes.DWORD),
        ("Data2", wintypes.WORD),
        ("Data3", wintypes.WORD),
        ("Data4", wintypes.BYTE * 8),
    ]

    def __init__(self, uuid_):
        ctypes.Structure.__init__(self)
        (
            self.Data1,
            self.Data2,
            self.Data3,
            self.Data4[0],
            self.Data4[1],
            rest,
        ) = uuid_.fields
        for i in range(2, 8):
            self.Data4[i] = rest >> (8 - i - 1) * 8 & 0xFF


class FOLDERID(object):  # [2]
    AccountPictures = UUID("{008ca0b1-55b4-4c56-b8a8-4de4b299d3be}")
    AdminTools = UUID("{724EF170-A42D-4FEF-9F26-B60E846FBA4F}")
    ApplicationShortcuts = UUID("{A3918781-E5F2-4890-B3D9-A7E54332328C}")
    CameraRoll = UUID("{AB5FB87B-7CE2-4F83-915D-550846C9537B}")
    CDBurning = UUID("{9E52AB10-F80D-49DF-ACB8-4330F5687855}")
    CommonAdminTools = UUID("{D0384E7D-BAC3-4797-8F14-CBA229B392B5}")
    CommonOEMLinks = UUID("{C1BAE2D0-10DF-4334-BEDD-7AA20B227A9D}")
    CommonPrograms = UUID("{0139D44E-6AFE-49F2-8690-3DAFCAE6FFB8}")
    CommonStartMenu = UUID("{A4115719-D62E-491D-AA7C-E74B8BE3B067}")
    CommonStartup = UUID("{82A5EA35-D9CD-47C5-9629-E15D2F714E6E}")
    CommonTemplates = UUID("{B94237E7-57AC-4347-9151-B08C6C32D1F7}")
    Contacts = UUID("{56784854-C6CB-462b-8169-88E350ACB882}")
    Cookies = UUID("{2B0F765D-C0E9-4171-908E-08A611B84FF6}")
    Desktop = UUID("{B4BFCC3A-DB2C-424C-B029-7FE99A87C641}")
    DeviceMetadataStore = UUID("{5CE4A5E9-E4EB-479D-B89F-130C02886155}")
    Documents = UUID("{FDD39AD0-238F-46AF-ADB4-6C85480369C7}")
    DocumentsLibrary = UUID("{7B0DB17D-9CD2-4A93-9733-46CC89022E7C}")
    Downloads = UUID("{374DE290-123F-4565-9164-39C4925E467B}")
    Favorites = UUID("{1777F761-68AD-4D8A-87BD-30B759FA33DD}")
    Fonts = UUID("{FD228CB7-AE11-4AE3-864C-16F3910AB8FE}")
    GameTasks = UUID("{054FAE61-4DD8-4787-80B6-090220C4B700}")
    History = UUID("{D9DC8A3B-B784-432E-A781-5A1130A75963}")
    ImplicitAppShortcuts = UUID("{BCB5256F-79F6-4CEE-B725-DC34E402FD46}")
    InternetCache = UUID("{352481E8-33BE-4251-BA85-6007CAEDCF9D}")
    Libraries = UUID("{1B3EA5DC-B587-4786-B4EF-BD1DC332AEAE}")
    Links = UUID("{bfb9d5e0-c6a9-404c-b2b2-ae6db6af4968}")
    LocalAppData = UUID("{F1B32785-6FBA-4FCF-9D55-7B8E7F157091}")
    LocalAppDataLow = UUID("{A520A1A4-1780-4FF6-BD18-167343C5AF16}")
    LocalizedResourcesDir = UUID("{2A00375E-224C-49DE-B8D1-440DF7EF3DDC}")
    Music = UUID("{4BD8D571-6D19-48D3-BE97-422220080E43}")
    MusicLibrary = UUID("{2112AB0A-C86A-4FFE-A368-0DE96E47012E}")
    NetHood = UUID("{C5ABBF53-E17F-4121-8900-86626FC2C973}")
    OriginalImages = UUID("{2C36C0AA-5812-4b87-BFD0-4CD0DFB19B39}")
    PhotoAlbums = UUID("{69D2CF90-FC33-4FB7-9A0C-EBB0F0FCB43C}")
    PicturesLibrary = UUID("{A990AE9F-A03B-4E80-94BC-9912D7504104}")
    Pictures = UUID("{33E28130-4E1E-4676-835A-98395C3BC3BB}")
    Playlists = UUID("{DE92C1C7-837F-4F69-A3BB-86E631204A23}")
    PrintHood = UUID("{9274BD8D-CFD1-41C3-B35E-B13F55A758F4}")
    Profile = UUID("{5E6C858F-0E22-4760-9AFE-EA3317B67173}")
    ProgramData = UUID("{62AB5D82-FDC1-4DC3-A9DD-070D1D495D97}")
    ProgramFiles = UUID("{905e63b6-c1bf-494e-b29c-65b732d3d21a}")
    ProgramFilesX64 = UUID("{6D809377-6AF0-444b-8957-A3773F02200E}")
    ProgramFilesX86 = UUID("{7C5A40EF-A0FB-4BFC-874A-C0F2E0B9FA8E}")
    ProgramFilesCommon = UUID("{F7F1ED05-9F6D-47A2-AAAE-29D317C6F066}")
    ProgramFilesCommonX64 = UUID("{6365D5A7-0F0D-45E5-87F6-0DA56B6A4F7D}")
    ProgramFilesCommonX86 = UUID("{DE974D24-D9C6-4D3E-BF91-F4455120B917}")
    Programs = UUID("{A77F5D77-2E2B-44C3-A6A2-ABA601054A51}")
    Public = UUID("{DFDF76A2-C82A-4D63-906A-5644AC457385}")
    PublicDesktop = UUID("{C4AA340D-F20F-4863-AFEF-F87EF2E6BA25}")
    PublicDocuments = UUID("{ED4824AF-DCE4-45A8-81E2-FC7965083634}")
    PublicDownloads = UUID("{3D644C9B-1FB8-4f30-9B45-F670235F79C0}")
    PublicGameTasks = UUID("{DEBF2536-E1A8-4c59-B6A2-414586476AEA}")
    PublicLibraries = UUID("{48DAF80B-E6CF-4F4E-B800-0E69D84EE384}")
    PublicMusic = UUID("{3214FAB5-9757-4298-BB61-92A9DEAA44FF}")
    PublicPictures = UUID("{B6EBFB86-6907-413C-9AF7-4FC2ABF07CC5}")
    PublicRingtones = UUID("{E555AB60-153B-4D17-9F04-A5FE99FC15EC}")
    PublicUserTiles = UUID("{0482af6c-08f1-4c34-8c90-e17ec98b1e17}")
    PublicVideos = UUID("{2400183A-6185-49FB-A2D8-4A392A602BA3}")
    QuickLaunch = UUID("{52a4f021-7b75-48a9-9f6b-4b87a210bc8f}")
    Recent = UUID("{AE50C081-EBD2-438A-8655-8A092E34987A}")
    RecordedTVLibrary = UUID("{1A6FDBA2-F42D-4358-A798-B74D745926C5}")
    ResourceDir = UUID("{8AD10C31-2ADB-4296-A8F7-E4701232C972}")
    Ringtones = UUID("{C870044B-F49E-4126-A9C3-B52A1FF411E8}")
    RoamingAppData = UUID("{3EB685DB-65F9-4CF6-A03A-E3EF65729F3D}")
    RoamedTileImages = UUID("{AAA8D5A5-F1D6-4259-BAA8-78E7EF60835E}")
    RoamingTiles = UUID("{00BCFC5A-ED94-4e48-96A1-3F6217F21990}")
    SampleMusic = UUID("{B250C668-F57D-4EE1-A63C-290EE7D1AA1F}")
    SamplePictures = UUID("{C4900540-2379-4C75-844B-64E6FAF8716B}")
    SamplePlaylists = UUID("{15CA69B3-30EE-49C1-ACE1-6B5EC372AFB5}")
    SampleVideos = UUID("{859EAD94-2E85-48AD-A71A-0969CB56A6CD}")
    SavedGames = UUID("{4C5C32FF-BB9D-43b0-B5B4-2D72E54EAAA4}")
    SavedSearches = UUID("{7d1d3a04-debb-4115-95cf-2f29da2920da}")
    Screenshots = UUID("{b7bede81-df94-4682-a7d8-57a52620b86f}")
    SearchHistory = UUID("{0D4C3DB6-03A3-462F-A0E6-08924C41B5D4}")
    SearchTemplates = UUID("{7E636BFE-DFA9-4D5E-B456-D7B39851D8A9}")
    SendTo = UUID("{8983036C-27C0-404B-8F08-102D10DCFD74}")
    SidebarDefaultParts = UUID("{7B396E54-9EC5-4300-BE0A-2482EBAE1A26}")
    SidebarParts = UUID("{A75D362E-50FC-4fb7-AC2C-A8BEAA314493}")
    SkyDrive = UUID("{A52BBA46-E9E1-435f-B3D9-28DAA648C0F6}")
    SkyDriveCameraRoll = UUID("{767E6811-49CB-4273-87C2-20F355E1085B}")
    SkyDriveDocuments = UUID("{24D89E24-2F19-4534-9DDE-6A6671FBB8FE}")
    SkyDrivePictures = UUID("{339719B5-8C47-4894-94C2-D8F77ADD44A6}")
    StartMenu = UUID("{625B53C3-AB48-4EC1-BA1F-A1EF4146FC19}")
    Startup = UUID("{B97D20BB-F46A-4C97-BA10-5E3608430854}")
    System = UUID("{1AC14E77-02E7-4E5D-B744-2EB1AE5198B7}")
    SystemX86 = UUID("{D65231B0-B2F1-4857-A4CE-A8E7C6EA7D27}")
    Templates = UUID("{A63293E8-664E-48DB-A079-DF759E0509F7}")
    UserPinned = UUID("{9E3995AB-1F9C-4F13-B827-48B24B6C7174}")
    UserProfiles = UUID("{0762D272-C50A-4BB0-A382-697DCD729B80}")
    UserProgramFiles = UUID("{5CD7AEE2-2219-4A67-B85D-6C9CE15660CB}")
    UserProgramFilesCommon = UUID("{BCBD3057-CA5C-4622-B42D-BC56DB0AE516}")
    Videos = UUID("{18989B1D-99B5-455B-841C-AB7C74E4DDFC}")
    VideosLibrary = UUID("{491E922F-5643-4AF4-A7EB-4E7A138D8174}")
    Windows = UUID("{F38BF404-1D43-42F2-9305-67DE0B28FC23}")


class UserHandle(object):  # [3]
    current = wintypes.HANDLE(0)
    common = wintypes.HANDLE(-1)


_CoTaskMemFree = windll.ole32.CoTaskMemFree  # [4]
_CoTaskMemFree.restype = None
_CoTaskMemFree.argtypes = [ctypes.c_void_p]

_SHGetKnownFolderPath = windll.shell32.SHGetKnownFolderPath  # [5] [3]
_SHGetKnownFolderPath.argtypes = [
    ctypes.POINTER(GUID),
    wintypes.DWORD,
    wintypes.HANDLE,
    ctypes.POINTER(ctypes.c_wchar_p),
]


class PathNotFoundException(Exception):
    pass


def get_path(folderid, user_handle=UserHandle.common):
    fid = GUID(folderid)
    pPath = ctypes.c_wchar_p()
    S_OK = 0
    if (
        _SHGetKnownFolderPath(ctypes.byref(fid), 0, user_handle, ctypes.byref(pPath))
        != S_OK
    ):
        raise PathNotFoundException()
    path = pPath.value
    _CoTaskMemFree(pPath)
    return path


if __name__ == "__main__":
    if len(sys.argv) < 2 or sys.argv[1] in ["-?", "/?"]:
        print("python knownpaths.py FOLDERID {current|common}")
        sys.exit(0)

    try:
        folderid = getattr(FOLDERID, sys.argv[1])
    except AttributeError:
        print('Unknown folder id "%s"' % sys.argv[1], file=sys.stderr)
        sys.exit(1)

    try:
        if len(sys.argv) == 2:
            print(get_path(folderid))
        else:
            print(get_path(folderid, getattr(UserHandle, sys.argv[2])))
    except PathNotFoundException:
        print('Folder not found "%s"' % " ".join(sys.argv[1:]), file=sys.stderr)
        sys.exit(1)

# [1] http://msdn.microsoft.com/en-us/library/windows/desktop/aa373931.aspx
# [2] http://msdn.microsoft.com/en-us/library/windows/desktop/dd378457.aspx
# [3] http://msdn.microsoft.com/en-us/library/windows/desktop/bb762188.aspx
# [4] http://msdn.microsoft.com/en-us/library/windows/desktop/ms680722.aspx
# [5] http://web.archive.org/web/20111025090317/http://www.themacaque.com/?p=954



# -*- coding: utf-8 -*-

import wx
import wx.xrc as xrc
from DisplayCAL.log import safe_print

try:
    from DisplayCAL.wxwindows import BetterStaticFancyText as StaticFancyText
except ImportError:
    from wx.lib.fancytext import StaticFancyText


class StaticFancyTextCtrlXmlHandler(xrc.XmlResourceHandler):
    def __init__(self):
        xrc.XmlResourceHandler.__init__(self)
        # Standard styles
        self.AddWindowStyles()

    def CanHandle(self, node):
        return self.IsOfClass(node, "StaticFancyText")

    # Process XML parameters and create the object
    def DoCreateResource(self):
        try:
            text = self.GetText("label")
        except Exception:
            text = ""
        w = StaticFancyText(
            self.GetParentAsWindow(),
            self.GetID(),
            text,
            pos=self.GetPosition(),
            size=self.GetSize(),
            style=self.GetStyle(),
            name=self.GetName(),
        )

        self.SetupWindow(w)
        if self.GetBool("hidden") and w.Shown:
            safe_print(f"{self.Name} should have been hidden")
            w.Hide()
        return w



# -*- coding: utf-8 -*-

try:
    from xml.etree import ElementTree as ET
except ImportError:
    pass


def dict2xml(d, elementname="element", pretty=True, allow_attributes=True, level=0):
    indent = pretty and "\t" * level or ""
    xml = []
    attributes = []
    children = []

    if isinstance(d, (dict, list)):
        start_tag = []
        start_tag.append(indent + "<" + elementname)

        if isinstance(d, dict):
            for key in d:
                value = d[key]
                if isinstance(value, (dict, list)) or not allow_attributes:
                    children.append(
                        dict2xml(value, key, pretty, allow_attributes, level + 1)
                    )
                else:
                    if pretty:
                        attributes.append("\n" + indent)
                    attributes.append(' %s="%s"' % (key, escape(str(value))))
        else:
            for value in d:
                children.append(
                    dict2xml(value, "item", pretty, allow_attributes, level + 1)
                )

        start_tag.extend(attributes)
        start_tag.append(children and ">" or "/>")
        xml.append("".join(start_tag))

        if children:
            for child in children:
                xml.append(child)

            xml.append("%s</%s>" % (indent, elementname))
    else:
        xml.append("%s<%s>%s</%s>" % (indent, elementname, escape(str(d)), elementname))

    return (pretty and "\n" or "").join(xml)


def escape(xml):
    return (
        xml.replace("&", "&amp;")
        .replace("<", "&lt;")
        .replace(">", "&gt;")
        .replace('"', "&quot;")
    )


class ETreeDict(dict):

    # Roughly follow "Converting Between XML and JSON"
    # https://www.xml.com/pub/a/2006/05/31/converting-between-xml-and-json.html

    def __init__(self, etree):
        super(ETreeDict, self).__init__()
        children = len(etree)
        if etree.attrib or etree.text or children:
            self[etree.tag] = dict(("@" + k, v) for k, v in etree.attrib.items())
            if etree.text:
                text = etree.text.strip()
                if etree.attrib or children:
                    if text:
                        self[etree.tag]["#text"] = text
                else:
                    self[etree.tag] = text
            if children:
                d = self[etree.tag]
                for child in etree:
                    for k, v in ETreeDict(child).items():
                        if k in d:
                            if not isinstance(d[k], list):
                                d[k] = [d[k]]
                            d[k].append(v)
                        else:
                            d[k] = v
        else:
            self[etree.tag] = None

    def __repr__(self):
        """od.__repr__() <==> repr(od)"""
        l = []
        for k in self:
            v = self[k]
            l.append("%r: %r" % (k, v))
        return "{%s}" % ", ".join(l)

    @property
    def json(self):
        import json

        return json.dumps(self)


class XMLDict(ETreeDict):
    def __init__(self, xml):
        etree = ET.fromstring(xml)
        ETreeDict.__init__(self, etree)



# -*- coding: utf-8 -*-
from __future__ import annotations

import sys
from unittest import mock

import pytest
from _pytest.monkeypatch import MonkeyPatch

from DisplayCAL import RealDisplaySizeMM, config
from DisplayCAL.dev.mocks import check_call
from tests.data.display_data import DisplayData

try:
    from tests.data.fake_dbus import FakeDBusObject
except ImportError:
    pass


def test_real_display_size_mm():
    """Test DisplayCAL.RealDisplaySizeMM.RealDisplaySizeMM() function."""
    RealDisplaySizeMM._displays = None
    assert RealDisplaySizeMM._displays is None
    with check_call(
        RealDisplaySizeMM, "_enumerate_displays", DisplayData.enumerate_displays()
    ):
        with check_call(config, "getcfg", DisplayData.CFG_DATA, call_count=2):
            display_size = RealDisplaySizeMM.RealDisplaySizeMM(0)
    assert display_size != (0, 0)
    assert display_size[0] > 1
    assert display_size[1] > 1


def test_xrandr_output_x_id_1():
    """Test DisplayCAL.RealDisplaySizeMM.GetXRandROutputXID() function."""
    RealDisplaySizeMM._displays = None
    assert RealDisplaySizeMM._displays is None
    with check_call(
        RealDisplaySizeMM, "_enumerate_displays", DisplayData.enumerate_displays()
    ):
        with check_call(config, "getcfg", DisplayData.CFG_DATA, call_count=2):
            result = RealDisplaySizeMM.GetXRandROutputXID(0)
    assert result != 0


def test_enumerate_displays():
    """Test DisplayCAL.RealDisplaySizeMM.enumerate_displays() function."""
    RealDisplaySizeMM._displays = None
    assert RealDisplaySizeMM._displays is None
    with check_call(
        RealDisplaySizeMM, "_enumerate_displays", DisplayData.enumerate_displays()
    ):
        result = RealDisplaySizeMM.enumerate_displays()
    assert result[0]["description"] != ""
    assert result[0]["edid"] != ""
    assert result[0]["icc_profile_atom_id"] != ""
    assert result[0]["icc_profile_output_atom_id"] != ""
    assert result[0]["name"] != ""
    assert result[0]["output"] != ""
    assert result[0]["pos"] != ""
    assert result[0]["ramdac_screen"] != ""
    assert result[0]["screen"] != ""
    assert result[0]["size"] != ""
    assert isinstance(result[0]["size"][0], int)
    assert isinstance(result[0]["size"][1], int)
    assert result[0]["size_mm"] != ""
    assert isinstance(result[0]["size_mm"][0], int)
    assert isinstance(result[0]["size_mm"][1], int)
    assert result[0]["x11_screen"] != ""
    # assert result[0]["xrandr_name"] != ""
    assert RealDisplaySizeMM._displays is not None


def test_get_display():
    """Test DisplayCAL.RealDisplaySizeMM.get_display() function."""
    RealDisplaySizeMM._displays = None
    assert RealDisplaySizeMM._displays is None
    with check_call(
        RealDisplaySizeMM, "_enumerate_displays", DisplayData.enumerate_displays()
    ):
        with check_call(config, "getcfg", DisplayData.CFG_DATA, call_count=2):
            display = RealDisplaySizeMM.get_display()
    assert RealDisplaySizeMM._displays is not None
    assert isinstance(display, dict)


def test_get_x_display():
    """Test DisplayCAL.RealDisplaySizeMM.get_x_display() function."""
    RealDisplaySizeMM._displays = None
    assert RealDisplaySizeMM._displays is None
    with check_call(
        RealDisplaySizeMM, "_enumerate_displays", DisplayData.enumerate_displays()
    ):
        with check_call(config, "getcfg", DisplayData.CFG_DATA, call_count=2):
            display = RealDisplaySizeMM.get_x_display(0)
    assert isinstance(display, tuple)
    assert len(display) == 3


@pytest.mark.parametrize(
    "function",
    (
        RealDisplaySizeMM.get_x_icc_profile_atom_id,
        RealDisplaySizeMM.get_x_icc_profile_output_atom_id,
    ),
)
def test_get_x_icc_profile_atom_id(function) -> None:
    """Test DisplayCAL.RealDisplaySizeMM.get_x_icc_profile_atom_id() function."""
    RealDisplaySizeMM._displays = None
    assert RealDisplaySizeMM._displays is None
    with check_call(
        RealDisplaySizeMM, "_enumerate_displays", DisplayData.enumerate_displays()
    ):
        with check_call(config, "getcfg", DisplayData.CFG_DATA, call_count=2):
            result = function(0)
    assert result is not None
    assert isinstance(result, int)


@pytest.mark.skipif("fake_dbus" not in sys.modules, reason="requires the DBus library")
def test_get_wayland_display(monkeypatch: MonkeyPatch) -> None:
    """Test if wayland display is returned."""
    with mock.patch.object(RealDisplaySizeMM, "DBusObject", new=FakeDBusObject):
        display = RealDisplaySizeMM.get_wayland_display(0, 0, 0, 0)
    assert display["xrandr_name"] == "DP-2"
    assert display["size_mm"] == (597, 336)
